@@ -15,7 +15,7 @@ public class ObjectType extends Type
     size = 4;
   }
 
-  ObjectType (String name)
+  public ObjectType (String name)
   {
     this_name = name;
     size = 4;
@@ -26,7 +26,12 @@ public class ObjectType extends Type
   final static int ADD_METHODS_DONE = 2;
   // A ClassType that we can expect to have a corresponding reflectClass.
   final static int EXISTING_CLASS = 4;
-  int flags;
+  /* */ public int flags;
+
+  public final boolean isExisting()
+  {
+    return (flags & EXISTING_CLASS) != 0;
+  }
 
   public final void setExisting(boolean existing)
   {
@@ -51,13 +56,29 @@ public class ObjectType extends Type
     try
       {
 	if (reflectClass == null)
-	  reflectClass = Class.forName(getInternalName().replace('/', '.'));
+	  {
+	    /* #ifdef JAVA2 */
+	    /* Specifies optional 'initialize' argument. */
+	    reflectClass = Class.forName(getInternalName().replace('/', '.'),
+					 false, getClass().getClassLoader());
+	    /* #endif */
+	    /* #ifndef JAVA2 */
+	    // reflectClass = Class.forName(getInternalName().replace('/', '.'));
+	    /* #endif */
+	  }
         flags |= EXISTING_CLASS;
       }
     catch (java.lang.ClassNotFoundException ex)
       {
         if ((flags & EXISTING_CLASS) != 0)
-          throw new RuntimeException("no such class: "+getName());
+          {
+	    RuntimeException rex
+              = new RuntimeException("no such class: "+getName());
+            /* #ifdef use:java.lang.Throwable.getCause */
+            rex.initCause(ex);
+            /* #endif */
+            throw rex;
+          }
       }
     return reflectClass;
   }
