@@ -39,7 +39,7 @@ public class defun extends Syntax implements Printable
 
     if (defs instanceof ModuleExp)
       decl.setCanRead(true);
-    st = tr.makePair(st, this, tr.makePair(p, decl, p.cdr));
+    st = Translator.makePair(st, this, Translator.makePair(p, decl, p.cdr));
     forms.addElement (st);
     return true;
   }
@@ -68,8 +68,8 @@ public class defun extends Syntax implements Printable
 	  {
 	    Pair p2 = (Pair) p1.cdr;
 	    LambdaExp lexp = new LambdaExp();
-	    lambdaSyntax.rewrite(lexp, p2.car, p2.cdr, tr);
-	    lexp.setName(name);
+	    lambdaSyntax.rewrite(lexp, p2.car, p2.cdr, tr, null);
+	    lexp.setSymbol(name);
 	    if (p2 instanceof PairWithPosition)
 	      {
 		PairWithPosition pp = (PairWithPosition) p2;
@@ -82,7 +82,7 @@ public class defun extends Syntax implements Printable
 	    sexp.setFuncDef(true);
 	    if (decl != null)
 	      {
-		sexp.binding = decl;
+		sexp.setBinding(decl);
 		if (decl.context instanceof ModuleExp && decl.getCanWrite())
 		  value = null;
 		decl.noteValue(value);
