@@ -11,9 +11,29 @@ public abstract class ScopeExp extends Expression
   Declaration decls;
   Declaration last;
 
-  Scope scope;
+  private Scope scope;
 
-  public final Declaration firstDecl () { return decls; }
+  public Declaration firstDecl () { return decls; }
+
+  public Scope getVarScope ()
+  {
+    Scope sc = scope;
+    if (sc == null)
+      scope = sc = new Scope();
+    return sc;
+  }
+
+  /** Clear bytecode resources for the ScopeExp.
+   * This potentially allows Kawa to generate code for the same (inlined,
+   * shared) ScopeExp multiple times - though we're not making use of that yet.
+   */
+  public void popScope (CodeAttr code)
+  {
+    for (Declaration decl = firstDecl(); decl != null; decl = decl.nextDecl())
+      decl.var = null;
+    code.popScope();
+    scope = null;
+  }
 
   public void add (Declaration decl)
   {
@@ -44,6 +64,27 @@ public abstract class ScopeExp extends Expression
     decl.context = this;
   }
 
+  /** Replace the <code>prev.next</code> by <code>newDecl</code>.
+   * If <code>prev==null</code>, replace the first decl. */
+  public void replaceFollowing (Declaration prev, Declaration newDecl)
+  {
+    Declaration oldDecl;
+    if (prev == null)
+      {
+	oldDecl = decls;
+	decls = newDecl;
+      }
+    else
+      {
+	oldDecl = prev.next;
+	prev.next = newDecl;
+      }
+    newDecl.next = oldDecl.next;
+    if (last == oldDecl)
+      last = newDecl;
+    oldDecl.next = null;
+  }
+
   public void remove (Declaration decl)
   {
     Declaration prev = null;
@@ -54,7 +95,7 @@ public abstract class ScopeExp extends Expression
 	    remove(prev, decl);
 	    return;
 	  }
-	prev = decl;
+	prev = cur;
       }
   }
 
@@ -68,7 +109,7 @@ public abstract class ScopeExp extends Expression
       last = prev;
   }
 
-  public ScopeExp () { scope = new Scope (); }
+  public ScopeExp () { }
 
   /** The statically enclosing binding contour. */
   public ScopeExp outer;
@@ -113,13 +154,13 @@ public abstract class ScopeExp extends Expression
     return null;
   }
 
-  public Declaration lookup (Object sym, Interpreter interp, int namespace)
+  public Declaration lookup (Object sym, Language language, int namespace)
   {
     for (Declaration decl = firstDecl();
          decl != null;  decl = decl.nextDecl())
       {
 	if (decl.symbol == sym
-	    && (interp.getNamespaceOf(decl) & namespace) != 0)
+	    && (language.getNamespaceOf(decl) & namespace) != 0)
 	  return decl;
       }
     return null;
@@ -148,33 +189,8 @@ public abstract class ScopeExp extends Expression
       decl.flags &= ~ (Declaration.NOT_DEFINING|Declaration.IS_UNKNOWN);
     else
       {
-	StringBuffer sbuf = new StringBuffer(200);
-	sbuf.append("duplicate definition for '");
-	sbuf.append(name);
-	int oldLine = decl.getLine();
-	if (oldLine <= 0)
-	  sbuf.append('\'');
-	else
-	  {
-	    sbuf.append("\' (overrides ");
-	    String oldFile = decl.getFile();
-	    if (oldFile == null || oldFile.equals(parser.getFile()))
-	      sbuf.append("line ");
-	    else
-	      {
-		sbuf.append(oldFile);
-		sbuf.append(':');
-	      }
-	    sbuf.append(oldLine);
-	    int oldColumn = decl.getColumn();
-	    if (oldColumn > 0)
-	      {
-		sbuf.append(':');
-		sbuf.append(oldColumn);
-	      }
-	    sbuf.append(')');
-	  }
-	parser.error(severity, sbuf.toString());
+        parser.error(severity, "duplicate definition of '"+name+"' here");
+        parser.error(severity, decl, "previous definition of '", "' here");
 	decl = addDeclaration(name);
       }
     return decl;
@@ -187,7 +203,7 @@ public abstract class ScopeExp extends Expression
   public final Declaration addDeclaration (Object name)
   {
     Declaration decl = new Declaration (name);
-    addDeclaration(decl);
+    add(decl);
     return decl;
   }
 
@@ -198,9 +214,8 @@ public abstract class ScopeExp extends Expression
    */
   public final Declaration addDeclaration (Object name, Type type)
   {
-    Declaration decl = new Declaration (name);
-    addDeclaration(decl);
-    decl.setType(type);
+    Declaration decl = new Declaration (name, type);
+    add(decl);
     return decl;
   }
 
@@ -220,8 +235,25 @@ public abstract class ScopeExp extends Expression
     return n;
   }
 
+  public static int nesting (ScopeExp sc)
+  {
+    int n = 0;
+    while (sc != null)
+      {
+	sc = sc.outer;
+	n++;
+      }
+    return n;
+  }
+
   protected Expression walk (ExpWalker walker)
   {
     return walker.walkScopeExp(this);
   }
+
+  public String toString() { return getClass().getName()+"#"+id; }
+
+  static int counter;
+  /** Unique id number, to ease print-outs and debugging. */
+  public int id = ++counter;
 }
