@@ -1,7 +1,6 @@
 package gnu.jemacs.buffer;
 import java.io.*;
 import gnu.mapping.*;
-import javax.swing.text.*;
 
 /** Insert input from an InputStream into a buffer.
  * This is typically output from an inferior process. */
@@ -34,7 +33,7 @@ class InputStreamHandler extends Thread
 	      break;
 	    wr.buffer = buffer;
 	    wr.count = avail;
-	    javax.swing.SwingUtilities.invokeAndWait(wr);
+	    mode.invoke(wr);
 	    //wr.write(buffer, 0, avail);
 	    //wr.flush();
 	  }
