@@ -16,6 +16,11 @@ public class lambda extends Lambda
 
   public void rewriteBody(LambdaExp lexp, Object body, Translator tr)
   {
+    tr.push(lexp);
+    if (lexp.defaultArgs != null)
+      for (int i = 0, n = lexp.defaultArgs.length;  i < n;  i++)
+	lexp.defaultArgs[i] = tr.rewrite(lexp.defaultArgs[i]);
+
     Pair pair;
     int i = 0;
     if (body instanceof Pair
@@ -64,7 +69,7 @@ public class lambda extends Lambda
 	  {
 	    Declaration decl = let.addDeclaration(arg.getSymbol());
 	    decl.setFluid(true);
-	    decl.setType(gnu.expr.FluidLetExp.typeFluidBinding);
+	    decl.setIndirectBinding(true);
 	    inits[i] = new ReferenceExp(arg);
 	    decl.noteValue(inits[i]);
 	  }
@@ -89,7 +94,7 @@ public class lambda extends Lambda
             else
               {
                 LambdaExp ilexp = new LambdaExp();
-                rewrite(ilexp, LList.Empty, interactive, tr);
+                rewrite(ilexp, LList.Empty, interactive, tr, null);
                 ilexp.setCanRead(true);
                 interactive = ilexp;
               }
