@@ -14,9 +14,10 @@ public class SlotGet extends Procedure2
   boolean isStatic;
 
   Procedure setter;
-  public static SlotGet field = new SlotGet("field", false, SlotSet.setField$Ex);
-  public static SlotGet staticField
-  = new SlotGet("static-field", true, SlotSet.setStaticField$Ex);
+  public static final SlotGet field
+    = new SlotGet("field", false, SlotSet.set$Mnfield$Ex);
+  public static final SlotGet staticField
+    = new SlotGet("static-field", true, SlotSet.set$Mnstatic$Mnfield$Ex);
 
   public SlotGet(String name, boolean isStatic)
   {
@@ -44,15 +45,17 @@ public class SlotGet extends Procedure2
   public Object apply2 (Object obj, Object name)
   {
     if (! (name instanceof String) && ! (name instanceof FString))
-      throw WrongType.make(null, this, 1);
-    Interpreter interpreter = Interpreter.getInterpreter();
+      throw WrongType.make(null, this, 2, name);
+    Language language = Language.getDefaultLanguage();
     String fname = gnu.expr.Compilation.mangleNameIfNeeded(name.toString());
     Class clas = isStatic ? coerceToClass(obj) : obj.getClass();
     if (clas.isArray() && "length".equals(fname))
       {
 	int length = java.lang.reflect.Array.getLength(obj);
-	return interpreter.coerceToObject(length);
+	return language.coerceToObject(length);
       }
+    if ("class".equals(fname))
+      return clas;
     boolean illegalAccess = false;
     java.lang.reflect.Field field;
     try
@@ -71,14 +74,7 @@ public class SlotGet extends Procedure2
                                      + fname + "' using `" + getName() + '\'');
         try
           {
-            Object result = field.get(obj);
-	    if (result instanceof Symbol
-		&& ((field.getModifiers() & java.lang.reflect.Modifier.FINAL)
-		    != 0))
-	      result = ((Symbol) result).get();
-	    else
-	      result = interpreter.coerceToObject(field.getType(), result);
-            return result;
+            return language.coerceToObject(field.getType(), field.get(obj));
           }
         catch (IllegalAccessException ex)
           {
@@ -89,18 +85,26 @@ public class SlotGet extends Procedure2
           }
       }
 
-    // Try looking for a method "getFname" instead:
-    String mname = ClassExp.slotToMethodName("get", fname);
+    // Try looking for a method "getFname" or "isFname" instead:
     try
       {
-        java.lang.reflect.Method getmethod
-          = clas.getMethod(mname, noClasses);
+        String mname = null;
+        java.lang.reflect.Method getmethod = null;
+        
+        try {
+          mname = ClassExp.slotToMethodName("get", fname);
+          getmethod = clas.getMethod(mname, noClasses);
+        } catch (Exception getEx) {
+          mname = ClassExp.slotToMethodName("is", fname);
+          getmethod = clas.getMethod(mname, noClasses);
+        }
+
         if (isStatic
             && (getmethod.getModifiers() & java.lang.reflect.Modifier.STATIC) == 0)
           throw new RuntimeException("cannot call non-static getter method `"
                                      + mname + "' using `" + getName() + '\'');
         Object result = getmethod.invoke(obj, Values.noArgs);
-        result = interpreter.coerceToObject(getmethod.getReturnType(), result);
+        result = language.coerceToObject(getmethod.getReturnType(), result);
         return result;
       }
     catch (java.lang.reflect.InvocationTargetException ex2)
@@ -148,12 +152,17 @@ public class SlotGet extends Procedure2
     SlotSet.apply(isStatic, obj, (String) name, value);
   }
 
+  /** Get a named property - field or 'get' accessor method.
+   * @param type the class type declaring the property.
+   * @param name the source (unmangled) name of the property.
+   */
   static Object getField(Type type, String name)
   {
     if (type instanceof ClassType && name != null)
       {
         ClassType clas = (ClassType) type;
-        gnu.bytecode.Field field = clas.getField(name);
+        gnu.bytecode.Field field
+          = clas.getField(Compilation.mangleNameIfNeeded(name));
         if (field != null)
           return field;
 
@@ -188,11 +197,21 @@ public class SlotGet extends Procedure2
     Expression arg1 = args[1];
     Type type = isStatic ? kawa.standard.Scheme.exp2Type(arg0)
       : arg0.getType();
-    String name = ClassMethods.checkName(arg1, true);
+    String name = ClassMethods.checkName(arg1);
+    if ("class".equals(name)
+        // Ideally we should also handle array types ...  FIXME
+        && type instanceof ClassType)
+      {
+        comp.loadClassRef(type.getName());
+        target.compileFromStack(comp, Type.java_lang_Class_type);
+        return;
+      }
     CodeAttr code = comp.getCode();
     if (type instanceof ClassType && name != null)
       {
 	ClassType ctype = (ClassType) type;
+	ClassType caller = comp.curClass != null ? comp.curClass
+	  : comp.mainClass;
         Object part = getField(ctype, name);
         if (part instanceof gnu.bytecode.Field)
           {
@@ -202,6 +221,9 @@ public class SlotGet extends Procedure2
             if (isStatic && ! isStaticField)
               comp.error('e', ("cannot access non-static field `" + name
                                + "' using `" + getName() + '\''));
+	    if (caller != null && ! caller.isAccessible(ctype, modifiers))
+	      comp.error('e', "field "+ctype.getName()+'.'+name
+			 +" is not accessible here");
             args[0].compile(comp,
                             isStaticField ? Target.Ignore
                             : Target.pushValue(ctype));
@@ -243,23 +265,23 @@ public class SlotGet extends Procedure2
             else
               code.emitGetField(field);
 	    Type ftype = field.getType();
-	    if ("gnu.mapping.Symbol".equals(ftype.getName())
-		&& (modifiers & Access.FINAL) != 0)
-	      code.emitInvokeVirtual(Compilation.getLocationMethod);
-	    Interpreter interpreter = Interpreter.getInterpreter();
+	    Language language = Language.getDefaultLanguage();
 	    Class fclass = ftype.getReflectClass();
 	    if (fclass != null)
-	      ftype = interpreter.getTypeFor(fclass);
+	      ftype = language.getTypeFor(fclass);
 	    target.compileFromStack(comp, ftype);
             return;
           }
         if (part instanceof gnu.bytecode.Method)
           {
             gnu.bytecode.Method method = (gnu.bytecode.Method) part;
+	    int modifiers = method.getModifiers();
             boolean isStaticMethod = method.getStaticFlag();
             if (isStatic && ! isStaticMethod)
               comp.error('e', "cannot call non-static getter method `"
                          + name + "' using `" + getName() + '\'');
+	    if (caller != null && ! caller.isAccessible(ctype, modifiers))
+	      comp.error('e', "method "+method +" is not accessible here");
             args[0].compile(comp,
                             isStaticMethod ? Target.Ignore
                             : Target.pushValue(ctype));
