@@ -1,4 +1,4 @@
-// Copyright (c) 1999  Per M.A. Bothner.
+// Copyright (c) 1999, 2004  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.mapping;
@@ -66,16 +66,6 @@ public abstract class MethodProc extends ProcedureN
     return atypes[index];
   }
 
-  /** Match the incoming arguments.
-   * @param ctx where to save the matched result on success
-   * @param args the incoming argument list
-   * @return non-negative if the match succeeded, else negative
-   */
-  public abstract int match (CallContext ctx, Object[] args);
-  // FUTURE:
-  // On success, vars has been initialized so vars.run() will work.
-  // public abstract int match(CallContext vars);
-
   /** Return code from match:  Unspecified failure. */
   public static final int NO_MATCH = -1;
 
@@ -91,46 +81,28 @@ public abstract class MethodProc extends ProcedureN
   public static final int NO_MATCH_AMBIGUOUS = 0xfff30000;
 
   /** Return code from match: Invalid argument type.
-   * In that case the lower half is the 0-origin index of the first
+   * In that case the lower half is the 1-origin index of the first
    * argument that does not match. */
   public static final int NO_MATCH_BAD_TYPE = 0xfff40000;
 
-  public Object match (Object[] args)
-  {
-    CallContext ctx = CallContext.getInstance();
-    return match(ctx, args) == 0 ? ctx : null;
-    // FUTURE:
-    // vars.setArgs(args);
-    // return match(vars) == 0 ? vars : null;
-  }
-
-  public abstract Object applyV(CallContext ctx) throws Throwable;
-  // FUTURE:
-  // ctx.run();
-
+  /** Helper method to throw an exception if a <code>matchX</code>
+   * method fails. */
   public static RuntimeException
   matchFailAsException(int code, Procedure proc, Object[] args)
   {
     int arg = (short) code;
     code &= 0xffff0000;
-    if (code == NO_MATCH_TOO_FEW_ARGS || code == NO_MATCH_TOO_MANY_ARGS)
-      return new WrongArguments(proc, args.length);
     if (code != NO_MATCH_BAD_TYPE)
-      arg = WrongType.ARG_UNKNOWN;
-    throw new WrongType(proc, arg, null);
+      return new WrongArguments(proc, args.length);
+    return new WrongType(proc, arg, args[arg-1]);
   }
 
   public Object applyN(Object[] args) throws Throwable
   {
     checkArgCount(this, args.length);
-    CallContext vars = CallContext.getInstance();
-    int err = match(vars, args);
-    if (err != 0)
-      throw matchFailAsException(err, this, args);
-    // FUTURE:
-    // vars.run();
-    // return vars.getResult();
-    return applyV(vars);
+    CallContext ctx = CallContext.getInstance();
+    checkN(args, ctx);
+    return ctx.runUntilValue();
   }
 
   /** Return the more specific of the arguments.
@@ -186,20 +158,57 @@ public abstract class MethodProc extends ProcedureN
   /** Return the index of the most specific method. */
   public static int mostSpecific(MethodProc[] procs, int length)
   {
-    MethodProc best = null;
-    if (length == 0)
-      return -1;
-    int result = 0;
-    best = procs[0];
+    if (length <= 1) // Handles length==0 and length==1.
+      return length - 1;
+    // best is non-null if there is a single most specific method.
+    MethodProc best = procs[0];
+    // This array (which is allocated lazily) is used if there is is a set
+    // of bestn methods none of which are more specific than the others.
+    MethodProc[] bests = null;
+    // If best==null, then the index of the most specific method;
+    // otherwise the active length of the bests array.
+    int bestn = 0;
+  outer:
     for (int i = 1;  i < length;  i++)
       {
         MethodProc method = procs[i];
-        best = mostSpecific(best, method);
-        if (best == null)
-          return -1;
-        else if (best == method)
-          result = i;
+	if (best != null)
+	  {
+	    MethodProc winner = mostSpecific(best, method);
+	    if (winner == null)
+	      {
+		if (bests == null)
+		  bests = new MethodProc[length];
+		bests[0] = best;
+		bests[1] = method;
+		bestn = 2;
+		best = null;
+	      }
+	    else if (winner == method)
+	      {
+		best = method;
+		bestn = i;
+	      }
+	  }
+	else
+	  {
+	    for (int j = 0;  j < bestn;  j++)
+	      {
+		MethodProc old = bests[j];
+		MethodProc winner = mostSpecific(old, method);
+		if (winner == old)
+		  continue outer;
+		if (winner == null)
+		  {
+		    bests[bestn++] = method;
+		    continue outer;
+		  }
+	      }
+	    // At this point method is more specific than bests[0..bestn-1].
+	    best = method;
+	    bestn = i;
+	  }
       }
-    return result;
+    return best == null ? -1 : bestn;
   }
 }
