@@ -1,8 +1,7 @@
 // FIXME - this probably belongs somewhere else!
 package gnu.mapping;
-
 import java.io.PrintWriter;
 
 public interface Printable {
-   void print(java.io.PrintWriter ps);
+   void print(PrintWriter ps);
 }
