@@ -2,6 +2,7 @@
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.math;
+import java.io.*;
 
 /** A complex number using rectangular (Cartesian) plain double values.
  * @author Per Bothner
@@ -10,11 +11,15 @@ package gnu.math;
  * Also see below for copyrights for functions taken from fdlib and f2c.
  */
 
-public class DComplex extends Complex
+public class DComplex extends Complex implements Externalizable
 { 
   double real;
   double imag;
 
+  public DComplex ()
+  {
+  }
+
   public DComplex (double real, double imag)
   {
     this.real = real;
@@ -121,7 +126,7 @@ public class DComplex extends Complex
 	    return new DComplex (real * y_re - imag * y_im,
 				 real * y_im + imag * y_re);
 	  }
-	return Complex.mul (this, yc);
+	return Complex.times(this, yc);
       }
     return ((Numeric)y).mulReversed(this);
   }
@@ -355,4 +360,21 @@ public class DComplex extends Complex
       }
     return w;
   }
+
+  /**
+   * @serialData Writes the real part, followed by the imaginary part.
+   *   Both are written as doubles (using writeDouble).
+   */
+  public void writeExternal(ObjectOutput out) throws IOException
+  {
+    out.writeDouble(real);
+    out.writeDouble(imag);
+  }
+
+  public void readExternal(ObjectInput in)
+    throws IOException, ClassNotFoundException
+  {
+    real = in.readDouble();
+    imag = in.readDouble();
+  }
 }
