@@ -1,11 +1,11 @@
-// Copyright (c) 1999  Per M.A. Bothner.
+// Copyright (c) 1999, 2004  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.text;
 import java.io.*;
 
 /**
- * Framework for implementing lexical scanners.
+ * Framework for implementing lexical scanners and parsers.
  * @author	Per Bothner
  */
 
@@ -25,6 +25,33 @@ public class Lexer extends Reader
     this.messages = messages;
   }
 
+  /** Enter a nested expression.
+   * This is used in interactive mode to control whether to continue
+   * past end of line, depending on whether the expression is incomplete.
+   * @param promptChar Used in prompt string to indicate type of nesting.
+   * @return The previous value of promptChar, to be passed to popNesting.
+   */
+  public char pushNesting (char promptChar)
+  {
+    nesting++;
+    LineBufferedReader port = getPort();
+    char save = port.readState;
+    port.readState = promptChar;
+    return save;
+  }
+
+  /** Exit a nested expression, reversing pushNesting
+   * @param save Saved values return by prior pushNesting
+   */
+  public void popNesting (char save)
+  {
+    LineBufferedReader port = getPort();
+    port.readState = save;
+    nesting--;
+  }
+
+  protected int nesting;
+
   public final LineBufferedReader getPort() { return port; }
 
   public void close() throws java.io.IOException
@@ -69,6 +96,23 @@ public class Lexer extends Reader
     port.unread_quick();
   }
 
+  /**
+   * Check if the next character matches a given character.
+   * @param ch The character to match against.
+   * @return if the character read matches
+   * On a match, the position is advanced following that character.
+   */
+  public boolean checkNext(char ch)
+      throws java.io.IOException
+  {
+    int r = port.read();
+    if (r == ch)
+	return true;
+    if (r >= 0)
+      port.unread_quick();
+    return false;
+  }
+
   protected void skip_quick() throws java.io.IOException
   {
     port.skip_quick();
@@ -245,4 +289,27 @@ public class Lexer extends Reader
     buffer[len] = (char) ch;
     tokenBufferLength = len + 1;
   }
+
+
+  private int saveTokenBufferLength = -1;
+
+  /** Start tentative parsing.  Must be followed by a reset. */
+  public void mark ()
+    throws java.io.IOException
+  {
+    if (saveTokenBufferLength >= 0)
+      throw new Error("internal error: recursive call to mark not allowed");
+    port.mark(Integer.MAX_VALUE);
+    saveTokenBufferLength = tokenBufferLength;
+  }
+
+  /** Stop tentative parsing.  Return to position where we called mark. */
+  public void reset ()
+    throws java.io.IOException
+  {
+    if (saveTokenBufferLength < 0)
+      throw new Error("internal error: reset called without prior mark");
+    port.reset();
+    saveTokenBufferLength = -1;
+  }
 }
