@@ -1,8 +1,10 @@
-// Copyright (c) 2001  Per M.A. Bothner.
+// Copyright (c) 2001, 2004  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.text;
 import java.io.*;
+import gnu.mapping.ThreadLocation;
+import gnu.lists.LList;
 
 /** A pretty printer.
  *
@@ -35,6 +37,14 @@ public class PrettyWriter extends java.io.Writer
 
   /** Line length we should format to. */
   int lineLength = 80;
+  int miserWidth = 40;
+
+  public static ThreadLocation lineLengthLoc
+    = new ThreadLocation("line-length");
+  public static ThreadLocation miserWidthLoc
+    = new ThreadLocation("miser-width");
+  public static ThreadLocation indentLoc
+    = new ThreadLocation("indent");
 
   public boolean isPrettyPrinting;
 
@@ -54,7 +64,8 @@ public class PrettyWriter extends java.io.Writer
 
   /** The column the first character in the buffer will appear in.
    * Normally zero, but if we end up with a very long line with no breaks in it
-   * we might have to output part of it. Then this will no longer be zero. */
+   * we might have to output part of it. Then this will no longer be zero.
+   * Ditto after emitting a prompt. */
   int bufferStartColumn;
   
   /** The line number we are currently on. Used for *print-lines* abrevs and
@@ -104,96 +115,106 @@ public class PrettyWriter extends java.io.Writer
    * to output the buffer. The length is stored in the logical block stack. */
   char[] suffix = new char[initialBufferSize];
 
-  // We have a queue of pending operations.  This is primarily stored
-  // in the circular buffer queueInts.  There are different kinds of
-  // operation types, and each operation can require a variable number
-  // of elements in the buffer, depending on the operation type.  Given
-  // an operation at 'index', the type operation type code is
-  // 'getQueueType(index)' (one of the QUEUED_OP_XXX_TYPE macros
-  // below), and the number of elements in the buffer is
-  // 'getQueueSize(index)' (one of the QUEUED_OP_XXX_SIZE values
-  // below).  You can think of the various QUEUED_OP_XXX_TYPEs as
-  // "sub-classes" of queued operations, but instead of creating
-  // actual Java objects, we allocate the objects' fields in the
-  // queueInts and QueueStrings arrays, to avoid expensive object
-  // allocation.  The special QUEUED_OP_NOP_TYPE is a used as a
-  // marker for when there isn't enough space in the rest of buffer,
-  // so we have to wrap around to the start.  The other QUEUED_OP_XXX
-  // macros are the offsets of the various "fields" relative to the
-  // start index.
-      
   static final int QUEUE_INIT_ALLOC_SIZE = 300; // FIXME
+
+  /** A queue of pending operations.
+   * This is primarily stored in the circular buffer queueInts.  There
+   * are different kinds of operation types, and each operation can
+   * require a variable number of elements in the buffer, depending on
+   * the operation type.  Given an operation at 'index', the type
+   * operation type code is 'getQueueType(index)' (one of the
+   * QITEM_XXX_TYPE macros below), and the number of elements in the
+   * buffer is 'getQueueSize(index)' (one of the QITEM_XXX_SIZE values
+   * below).  You can think of the various QITEM_XXX_TYPEs as
+   * "sub-classes" of queued operations, but instead of creating
+   * actual Java objects, we allocate the objects' fields in the
+   * queueInts and QueueStrings arrays, to avoid expensive object
+   * allocation.  The special QITEM_NOP_TYPE is a used as a marker for
+   * when there isn't enough space in the rest of buffer, so we have
+   * to wrap around to the start.  The other QITEM_XXX macros are the
+   * offsets of the various "fields" relative to the start index. */
   int[] queueInts = new int[QUEUE_INIT_ALLOC_SIZE];
+
   /** For simplicity, queueStrings is the same size as queueInts. */
   String[] queueStrings = new String[QUEUE_INIT_ALLOC_SIZE];
   /** Index in queueInts and queueStrings of oldest enqueued operation. */
   int queueTail;
   /** Number of elements (in queueInts and queueStrings) in use. */
   int queueSize;
-  /** Index (into queueInts) of current unclosed begin-block node. */
-  int currentBlock;
+  /** If >= 0, index (into queueInts) of current unclosed begin-block node.
+   * This is a head of a linked linked of queued BLOCK_START for which
+   * we haven't seen the matching BLOCK_END  */
+  int currentBlock = -1;
+  /** Number of startLogicalBlock - number of endLogicalBlock. */
   public int pendingBlocksCount;
 
-  static final int QUEUED_OP_TYPE = 0;
+  /** The first it QITEM contains it type code and size.
+   * The type code is one of the QITEM_XXX_TYPE values below.
+   * The size is the corresponding QITEM_XXX_SIZE value below,
+   * except for the case of QITEM_NOP_TYPE (which is used as a filler). */
+  static final int QITEM_TYPE_AND_SIZE = 0;
   private int getQueueType(int index) { return queueInts[index] & 0xFF; }
   private int getQueueSize(int index) { return queueInts[index] >> 16; }
-  /** Relative offset of POSN field of a QUEUED_OP> */
-  static final int QUEUED_OP_POSN = 1;
-  /** Size of "base part" of a QUEUED_OP. */
-  static final int QUEUED_OP_SIZE = 2;
+  /** Relative offset of POSN field of a QITEM> */
+  static final int QITEM_POSN = 1;
+  /** Size of "base part" of a QITEM. */
+  static final int QITEM_BASE_SIZE = 2;
 
   /** A dummy queue item used at the high end of the queue buffer
    * when there isn't enough space for the needed queue item. */
-  static final int QUEUED_OP_NOP_TYPE = 0;
+  static final int QITEM_NOP_TYPE = 0;
 
   /** "Abstract" type for beginning of section.
    * A section is from a block-start to a newline, from a newline to
    * the next newline (in the same block?), or from a newline to
    * the block end (?). */
-  /*static final int QUEUED_OP_SECTION_START_TYPE = 1;*/
-  static final int QUEUED_OP_SECTION_START_SIZE = QUEUED_OP_SIZE + 2;
-  static final int QUEUED_OP_SECTION_START_DEPTH = QUEUED_OP_SIZE;
-  static final int QUEUED_OP_SECTION_START_SECTION_END = QUEUED_OP_SIZE + 1;
+  /*static final int QITEM_SECTION_START_TYPE = 1;*/
+  static final int QITEM_SECTION_START_SIZE = QITEM_BASE_SIZE + 2;
+  static final int QITEM_SECTION_START_DEPTH = QITEM_BASE_SIZE;
+  static final int QITEM_SECTION_START_SECTION_END = QITEM_BASE_SIZE + 1;
 
   /** A newline queue item. */
-  static final int QUEUED_OP_NEWLINE_TYPE = 2;
-  static final int QUEUED_OP_NEWLINE_SIZE = QUEUED_OP_SECTION_START_SIZE + 1;
-  static final int QUEUED_OP_NEWLINE_KIND = QUEUED_OP_SECTION_START_SIZE;
+  static final int QITEM_NEWLINE_TYPE = 2;
+  static final int QITEM_NEWLINE_SIZE = QITEM_SECTION_START_SIZE + 1;
+  static final int QITEM_NEWLINE_KIND = QITEM_SECTION_START_SIZE;
   public static final int NEWLINE_LINEAR = 'N';
   public static final int NEWLINE_LITERAL = 'L';
   public static final int NEWLINE_FILL = 'F';
+  /** A non-nested ' ' gets an implicit NEWLINE_SPACE.
+   * This is treated similarly to NEWLINE_FILL, but not quite. */
+  public static final int NEWLINE_SPACE = 'S';
   public static final int NEWLINE_MISER = 'M';
   public static final int NEWLINE_MANDATORY = 'R';  // "required"
 
-  static final int QUEUED_OP_INDENTATION_TYPE = 3;
-  static final int QUEUED_OP_INDENTATION_SIZE = QUEUED_OP_SIZE + 2;
-  static final int QUEUED_OP_INDENTATION_KIND = QUEUED_OP_SIZE;
-  static final int QUEUED_OP_INDENTATION_BLOCK = 'B';
-  static final int QUEUED_OP_INDENTATION_CURRENT = 'C';
-  static final int QUEUED_OP_INDENTATION_AMOUNT = QUEUED_OP_SIZE + 1;
+  static final int QITEM_INDENTATION_TYPE = 3;
+  static final int QITEM_INDENTATION_SIZE = QITEM_BASE_SIZE + 2;
+  static final int QITEM_INDENTATION_KIND = QITEM_BASE_SIZE;
+  static final char QITEM_INDENTATION_BLOCK = 'B';
+  static final char QITEM_INDENTATION_CURRENT = 'C';
+  static final int QITEM_INDENTATION_AMOUNT = QITEM_BASE_SIZE + 1;
 
   /** A "block-start" queue item. */
-  static final int QUEUED_OP_BLOCK_START_TYPE = 4;
-  static final int QUEUED_OP_BLOCK_START_SIZE = QUEUED_OP_SECTION_START_SIZE + 1;
-  /** If the QUEUED_OP_SECTION_START_SECTION_END < 0, it points to
+  static final int QITEM_BLOCK_START_TYPE = 4;
+  static final int QITEM_BLOCK_START_SIZE = QITEM_SECTION_START_SIZE + 3;
+  /** If the QITEM_BLOCK_START_BLOCK_END < 0, it points to
    * the previous (outer) un-closed block-start.
-   * If QUEUED_OP_SECTION_START_SECTION_END > 0, it points to the
+   * If QITEM_BLOCK_START_BLOCK_END > 0, it points to the
    * corresponding block-end node.
    * In both cases the pointers are relative to the current BLOCK_START. */
-  static final int QUEUED_OP_BLOCK_START_BLOCK_END = QUEUED_OP_SECTION_START_SIZE;
-  static final int QUEUED_OP_BLOCK_PREFIX = 0;
-  static final int QUEUED_OP_BLOCK_START_SUFFIX = 1;
+  static final int QITEM_BLOCK_START_BLOCK_END = QITEM_SECTION_START_SIZE;
+  static final int QITEM_BLOCK_START_PREFIX = QITEM_SECTION_START_SIZE + 1;
+  static final int QITEM_BLOCK_START_SUFFIX = QITEM_SECTION_START_SIZE + 2;
 
-  static final int QUEUED_OP_BLOCK_END_TYPE = 5;
-  static final int QUEUED_OP_BLOCK_END_SIZE = QUEUED_OP_SIZE;
+  static final int QITEM_BLOCK_END_TYPE = 5;
+  static final int QITEM_BLOCK_END_SIZE = QITEM_BASE_SIZE;
 
-  static final int QUEUED_OP_TAB_TYPE = 6;
-  static final int QUEUED_OP_TAB_SIZE = QUEUED_OP_SIZE + 3;
-  static final int QUEUED_OP_TAB_FLAGS = QUEUED_OP_SIZE;
-  static final int QUEUED_OP_TAB_IS_SECTION = 1;
-  static final int QUEUED_OP_TAB_IS_RELATIVE = 2;
-  static final int QUEUED_OP_TAB_COLNUM = QUEUED_OP_SIZE + 1;
-  static final int QUEUED_OP_TAB_COLINC = QUEUED_OP_SIZE + 2;
+  static final int QITEM_TAB_TYPE = 6;
+  static final int QITEM_TAB_SIZE = QITEM_BASE_SIZE + 3;
+  static final int QITEM_TAB_FLAGS = QITEM_BASE_SIZE;
+  static final int QITEM_TAB_IS_SECTION = 1;
+  static final int QITEM_TAB_IS_RELATIVE = 2;
+  static final int QITEM_TAB_COLNUM = QITEM_BASE_SIZE + 1;
+  static final int QITEM_TAB_COLINC = QITEM_BASE_SIZE + 2;
 
   private int getSectionColumn()
   {
@@ -227,7 +248,7 @@ public class PrettyWriter extends java.io.Writer
 
   public void write (int ch)
   {
-    //System.err.print("{WRITE-ch: "+((char)ch)+"}");
+    //log("{WRITE-ch: "+((char)ch)+"}");
     if (ch == '\n' && isPrettyPrinting)
       enqueueNewline(NEWLINE_LITERAL);
     else
@@ -236,6 +257,8 @@ public class PrettyWriter extends java.io.Writer
 	int fillPointer = bufferFillPointer;
 	buffer[fillPointer] = (char) ch;
 	bufferFillPointer = 1 + fillPointer;
+	if (ch == ' ' && isPrettyPrinting && currentBlock < 0)
+	  enqueueNewline(NEWLINE_SPACE);
       }
   }
 
@@ -246,7 +269,7 @@ public class PrettyWriter extends java.io.Writer
 
   public void write (String str, int start, int count)
   {
-    //System.err.print("{WRITE-str: "+str.substring(start, start+count)+"}");
+    //log("{WRITE-str: "+str.substring(start, start+count)+"}");
     while (count > 0)
       {
 	int cnt = count;
@@ -266,7 +289,15 @@ public class PrettyWriter extends java.io.Writer
 		fillPointer = bufferFillPointer;
 	      }
 	    else
-	      buffer[fillPointer++] = (char) ch;
+	      {
+		buffer[fillPointer++] = (char) ch;
+		if (ch == ' ' && isPrettyPrinting && currentBlock < 0)
+		  {
+		    bufferFillPointer = fillPointer;
+		    enqueueNewline(NEWLINE_SPACE);
+		    fillPointer = bufferFillPointer;
+		  }
+	      }
 	  }
 	bufferFillPointer = fillPointer;
       }
@@ -279,7 +310,7 @@ public class PrettyWriter extends java.io.Writer
 
   public void write (char[] str, int start, int count)
   {
-    //System.err.print("{WRITE: "+new String(str, start, count)+"}");
+    //log("{WRITE: "+new String(str, start, count)+"}");
     int end = start + count;
   retry:
     while (count > 0)
@@ -287,10 +318,13 @@ public class PrettyWriter extends java.io.Writer
 	// Look for newline.  Should be merged with following loop.  FIXME.
 	for (int i = start;  i < end;  i++)
 	  {
-	    if (str[i] == '\n' && isPrettyPrinting)
+	    char c;
+	    if (isPrettyPrinting
+		&& ((c = str[i]) == '\n'
+		    || (c == ' ' && currentBlock < 0)))
 	      {
-		write (str, start, i - start); // Recurse
-		enqueueNewline(NEWLINE_LITERAL);
+		write(str, start, i - start); // Recurse
+		write(c);
 		start = i + 1;
 		count = end - start;
 		continue retry;
@@ -375,10 +409,10 @@ public class PrettyWriter extends java.io.Writer
 
   int enqueueTab (int flags, int colnum, int colinc) // DONE
   {
-    int addr = enqueue(QUEUED_OP_TAB_TYPE, QUEUED_OP_TAB_SIZE);
-    queueInts[addr + QUEUED_OP_TAB_FLAGS] = flags;
-    queueInts[addr + QUEUED_OP_TAB_COLNUM] = colnum;
-    queueInts[addr + QUEUED_OP_TAB_COLINC] = colinc;
+    int addr = enqueue(QITEM_TAB_TYPE, QITEM_TAB_SIZE);
+    queueInts[addr + QITEM_TAB_FLAGS] = flags;
+    queueInts[addr + QITEM_TAB_COLNUM] = colnum;
+    queueInts[addr + QITEM_TAB_COLINC] = colinc;
     return addr;
   }
 
@@ -432,7 +466,7 @@ public class PrettyWriter extends java.io.Writer
     int oldLength = queueInts.length;
     int endAvail = oldLength - queueTail - queueSize;
     if (endAvail > 0 && size > endAvail)
-      enqueue (QUEUED_OP_NOP_TYPE, endAvail);
+      enqueue(QITEM_NOP_TYPE, endAvail);
     if (queueSize + size > oldLength)
       {
 	int newLength = enoughSpace(oldLength, size);
@@ -440,7 +474,7 @@ public class PrettyWriter extends java.io.Writer
 	String[] newStrings = new String[newLength];
 	int queueHead = queueTail + queueSize - oldLength;
 	if (queueHead > 0)
-	  {
+	  { // Wraps around.
 	    System.arraycopy(queueInts, 0, newInts, 0, queueHead);
 	    System.arraycopy(queueStrings, 0, newStrings, 0, queueHead);
 	  }
@@ -461,20 +495,22 @@ public class PrettyWriter extends java.io.Writer
     int addr = queueTail + queueSize;
     if (addr >= queueInts.length)
       addr -= queueInts.length;
-    queueInts[addr + QUEUED_OP_TYPE] = kind | (size << 16);
+    queueInts[addr + QITEM_TYPE_AND_SIZE] = kind | (size << 16);
     if (size > 1)
-      queueInts[addr + QUEUED_OP_POSN] = indexPosn(bufferFillPointer);
+      queueInts[addr + QITEM_POSN] = indexPosn(bufferFillPointer);
+    //log("enqueue "+itemKindString(kind)+" size:"+size+" at:"+queueSize+enqueueExtraLog); enqueueExtraLog = "";
     queueSize += size;
     return addr;
   }
 
-  public void enqueueNewline (int kind) // DONE
+  public void enqueueNewline (int kind)
   {
     int depth = pendingBlocksCount;
-    int newline = enqueue(QUEUED_OP_NEWLINE_TYPE, QUEUED_OP_NEWLINE_SIZE);
-    queueInts[newline + QUEUED_OP_NEWLINE_KIND] = kind;
-    queueInts[newline + QUEUED_OP_SECTION_START_DEPTH] = pendingBlocksCount;
-    queueInts[newline + QUEUED_OP_SECTION_START_SECTION_END] = 0;
+    //enqueueExtraLog = " kind:"+(char) kind;
+    int newline = enqueue(QITEM_NEWLINE_TYPE, QITEM_NEWLINE_SIZE);
+    queueInts[newline + QITEM_NEWLINE_KIND] = kind;
+    queueInts[newline + QITEM_SECTION_START_DEPTH] = pendingBlocksCount;
+    queueInts[newline + QITEM_SECTION_START_SECTION_END] = 0;
     int entry = queueTail;
     int todo = queueSize;
     while (todo > 0)
@@ -484,15 +520,15 @@ public class PrettyWriter extends java.io.Writer
 	if (entry == newline)
 	  break;
 	int type = getQueueType(entry);
-	if ((type == QUEUED_OP_NEWLINE_TYPE
-	     || type == QUEUED_OP_BLOCK_START_TYPE)
-	    && queueInts[entry + QUEUED_OP_SECTION_START_SECTION_END] == 0
-	    && depth <= queueInts[entry + QUEUED_OP_SECTION_START_DEPTH])
+	if ((type == QITEM_NEWLINE_TYPE
+	     || type == QITEM_BLOCK_START_TYPE)
+	    && queueInts[entry + QITEM_SECTION_START_SECTION_END] == 0
+	    && depth <= queueInts[entry + QITEM_SECTION_START_DEPTH])
 	  {
 	    int delta = newline - entry;
 	    if (delta < 0)
 	      delta += queueInts.length;
-	    queueInts[entry + QUEUED_OP_SECTION_START_SECTION_END] = delta;
+	    queueInts[entry + QITEM_SECTION_START_SECTION_END] = delta;
 	  }
 	int size = getQueueSize(entry);
 	todo -= size;
@@ -508,47 +544,76 @@ public class PrettyWriter extends java.io.Writer
       enqueueNewline(kind);
   }
 
-  public int enqueueIndent (int kind, int amount)
+  public int enqueueIndent (char kind, int amount)
   {
-    int result = enqueue(QUEUED_OP_INDENTATION_TYPE, QUEUED_OP_INDENTATION_SIZE);
-    queueInts[result + QUEUED_OP_INDENTATION_KIND] = kind;
-    queueInts[result + QUEUED_OP_INDENTATION_AMOUNT] = amount;
+    //enqueueExtraLog = " kind:"+kind+" amount:"+amount;
+    int result = enqueue(QITEM_INDENTATION_TYPE, QITEM_INDENTATION_SIZE);
+    queueInts[result + QITEM_INDENTATION_KIND] = kind;
+    queueInts[result + QITEM_INDENTATION_AMOUNT] = amount;
     return result;
   }
 
   public void addIndentation(int amount, boolean current)
   {
     if (isPrettyPrinting)
-      enqueueIndent((current ? QUEUED_OP_INDENTATION_CURRENT
-		     : QUEUED_OP_INDENTATION_BLOCK),
+      enqueueIndent((current ? QITEM_INDENTATION_CURRENT
+		     : QITEM_INDENTATION_BLOCK),
 		    amount);
   }
 
   public void startLogicalBlock (String prefix, boolean perLine, String suffix)
   {
+    // If the queue is empty, it is a good time to check if line-length etc
+    // have been changed.
+    if (queueSize == 0 && bufferFillPointer == 0)
+      {
+        Object llen = lineLengthLoc.get(null);
+        if (llen == null)
+          lineLength = 80;
+        else
+          lineLength = Integer.parseInt(llen.toString());
+
+        Object mwidth = miserWidthLoc.get(null);
+        if (mwidth == null || mwidth == Boolean.FALSE
+            // For Common Lisp nil.  Should we use Language.isTrue() FIXME.
+            || mwidth == LList.Empty)
+          miserWidth = -1;
+        else
+          miserWidth = Integer.parseInt(mwidth.toString());
+
+        Object indent = indentLoc.get(null);
+        // if (indent == null || indent ...
+      }
     if (prefix != null)
       write(prefix);
     if (! isPrettyPrinting)
       return;
-    int start = enqueue (QUEUED_OP_BLOCK_START_TYPE,
-			 QUEUED_OP_BLOCK_START_SIZE);
-    queueInts[start + QUEUED_OP_SECTION_START_DEPTH] = pendingBlocksCount;
-    queueStrings[start + QUEUED_OP_BLOCK_PREFIX] = perLine ? prefix : null;
-    queueStrings[start + QUEUED_OP_BLOCK_START_SUFFIX] = suffix;
+    int start = enqueue (QITEM_BLOCK_START_TYPE,
+			 QITEM_BLOCK_START_SIZE);
+    queueInts[start + QITEM_SECTION_START_DEPTH] = pendingBlocksCount;
+    queueStrings[start + QITEM_BLOCK_START_PREFIX]
+      = perLine ? prefix : null;
+    queueStrings[start + QITEM_BLOCK_START_SUFFIX] = suffix;
     pendingBlocksCount++;
-    currentBlock -= start;
-    if (currentBlock > 0)
-      currentBlock -= queueInts.length;
-    queueInts[start + QUEUED_OP_SECTION_START_SECTION_END] = 0;
-    queueInts[start + QUEUED_OP_BLOCK_START_BLOCK_END] = currentBlock;
+    int outerBlock = currentBlock;
+    if (outerBlock < 0)
+      outerBlock = 0;
+    else
+      {
+	outerBlock -= start;
+	if (outerBlock > 0)
+	  outerBlock -= queueInts.length;
+      }
+    queueInts[start + QITEM_BLOCK_START_BLOCK_END] = outerBlock;
+    queueInts[start + QITEM_SECTION_START_SECTION_END] = 0;
     currentBlock = start;
   }
 
   public void endLogicalBlock ()
   {
-    int end = enqueue (QUEUED_OP_BLOCK_END_TYPE, QUEUED_OP_BLOCK_END_SIZE);
+    int end = enqueue (QITEM_BLOCK_END_TYPE, QITEM_BLOCK_END_SIZE);
     pendingBlocksCount--;
-    if (blockDepth >= LOGICAL_BLOCK_LENGTH * (pendingBlocksCount + 2))
+    if (currentBlock < 0)
       {
 	// reallyStartLogicalBlock has been called for the matching
 	// BEGIN_BLOCK, so it is no longer in the queue.  Instead it is in
@@ -560,21 +625,41 @@ public class PrettyWriter extends java.io.Writer
 	  write(this.suffix,
 		this.suffix.length - suffixLength,
 		suffixLength - suffixPreviousLength);
+	currentBlock = -1;
 	return;
       }
     int start = currentBlock;
-    currentBlock = queueInts[start + QUEUED_OP_BLOCK_START_BLOCK_END];
-    // Make currentBlock absolute instead of relative.
-    currentBlock += start;
-    if (currentBlock < 0)
-      currentBlock += queueInts.length;
-    String suffix = queueStrings[start + QUEUED_OP_BLOCK_START_SUFFIX];
+    int outerBlock = queueInts[start + QITEM_BLOCK_START_BLOCK_END];
+    if (outerBlock == 0)
+      currentBlock = -1;
+    else
+      {
+	int qtailFromStart = queueTail - start;
+	if (qtailFromStart > 0)
+	  qtailFromStart -= queueInts.length;
+	if (outerBlock < qtailFromStart)
+	  {
+	    // reallyStartLogicalBlock has been called for the outer block,
+	    // so there is no currentBlock.
+	    currentBlock = -1;
+	  }
+	else
+	  {
+	    // Make currentBlock absolute instead of relative.
+	    outerBlock += start;
+	    if (outerBlock < 0)
+	      outerBlock += queueInts.length;
+	    currentBlock = outerBlock;
+	  }
+      }
+    String suffix = queueStrings[start + QITEM_BLOCK_START_SUFFIX];
     if (suffix != null)
       write(suffix);
     int endFromStart = end - start;
     if (endFromStart < 0) // wrap-around.
       endFromStart += queueInts.length;
-    queueInts[start + QUEUED_OP_BLOCK_START_BLOCK_END] = endFromStart;
+    queueInts[start + QITEM_BLOCK_START_BLOCK_END] = endFromStart;
+    //log("endLogicalBlock end:"+end+" start:"+start+" rel:"+endFromStart);
   }
 
   public void endLogicalBlock (String suffix)
@@ -589,12 +674,12 @@ public class PrettyWriter extends java.io.Writer
 
   int computeTabSize (int tab, int sectionStart, int column) // DONE
   {
-    int flags = queueInts[tab + QUEUED_OP_TAB_FLAGS];
-    boolean isSection = (flags & QUEUED_OP_TAB_IS_SECTION) != 0;
-    boolean isRelative = (flags & QUEUED_OP_TAB_IS_RELATIVE) != 0;
+    int flags = queueInts[tab + QITEM_TAB_FLAGS];
+    boolean isSection = (flags & QITEM_TAB_IS_SECTION) != 0;
+    boolean isRelative = (flags & QITEM_TAB_IS_RELATIVE) != 0;
     int origin = isSection ? sectionStart : 0;
-    int colnum = queueInts[tab + QUEUED_OP_TAB_COLNUM];
-    int colinc = queueInts[tab + QUEUED_OP_TAB_COLINC];
+    int colnum = queueInts[tab + QITEM_TAB_COLNUM];
+    int colinc = queueInts[tab + QITEM_TAB_COLINC];
     if (isRelative)
       {
 	if (colinc > 1)
@@ -625,18 +710,18 @@ public class PrettyWriter extends java.io.Writer
 	if (op >= queueInts.length)
 	  op = 0;
 	int type = getQueueType(op);
-	if (type != QUEUED_OP_NOP_TYPE)
+	if (type != QITEM_NOP_TYPE)
 	  {
-	    int posn = queueInts[op + QUEUED_OP_POSN];
+	    int posn = queueInts[op + QITEM_POSN];
 	    if (posn >= endPosn)
 	      break;
-	    if (type == QUEUED_OP_TAB_TYPE)
+	    if (type == QITEM_TAB_TYPE)
 	      column += computeTabSize(op, sectionStart,
 				       column + posnIndex (posn));
-	    else if (type == QUEUED_OP_NEWLINE_TYPE
-		     || type == QUEUED_OP_BLOCK_START_TYPE)
+	    else if (type == QITEM_NEWLINE_TYPE
+		     || type == QITEM_BLOCK_START_TYPE)
 	      sectionStart
-		= column + posnIndex(queueInts[op + QUEUED_OP_POSN]);
+		= column + posnIndex(queueInts[op + QITEM_POSN]);
 	  }
 	int size = getQueueSize(op);
 	todo -= size;
@@ -645,7 +730,7 @@ public class PrettyWriter extends java.io.Writer
     return column + index;
   }
 
-  void expandTabs (int through) // ODNE except FIXMEs
+  void expandTabs (int through)
   {
     int numInsertions = 0;
     int additional = 0;
@@ -661,9 +746,9 @@ public class PrettyWriter extends java.io.Writer
 	if (op == through)
 	  break;
 	int type = getQueueType(op);
-	if (type == QUEUED_OP_TAB_TYPE)
+	if (type == QITEM_TAB_TYPE)
 	  {
-	    int index = posnIndex(queueInts[op + QUEUED_OP_POSN]);
+	    int index = posnIndex(queueInts[op + QITEM_POSN]);
 	    int tabsize = computeTabSize (op, sectionStart, column + index);
 	    if (tabsize != 0)
 	      {
@@ -681,9 +766,9 @@ public class PrettyWriter extends java.io.Writer
 		column += tabsize;
 	      }
 	  }
-	else if (op == QUEUED_OP_NEWLINE_TYPE || op == QUEUED_OP_BLOCK_START_TYPE)
+	else if (op == QITEM_NEWLINE_TYPE || op == QITEM_BLOCK_START_TYPE)
 	  {
-	    sectionStart = column + posnIndex(queueInts[op + QUEUED_OP_POSN]);
+	    sectionStart = column + posnIndex(queueInts[op + QITEM_POSN]);
 	  }
 	int size = getQueueSize(op);
 	todo -= size;
@@ -751,8 +836,7 @@ public class PrettyWriter extends java.io.Writer
   boolean maybeOutput(boolean forceNewlines) // DONE
   {
     boolean outputAnything = false;
-    //System.err.print("maybeOutput("+forceNewlines+"):");  System.err.flush();
-    //dumpQueue();
+    //log("maybeOutput("+forceNewlines+"):");  dumpQueue();
   loop:
     while (queueSize > 0)
       {
@@ -762,9 +846,9 @@ public class PrettyWriter extends java.io.Writer
 	int type = getQueueType(next);
 	switch (type)
 	  {
-	  case QUEUED_OP_NEWLINE_TYPE:
+	  case QITEM_NEWLINE_TYPE:
 	    boolean cond;
-	    switch (queueInts[next+QUEUED_OP_NEWLINE_KIND])
+	    switch (queueInts[next+QITEM_NEWLINE_KIND])
 	      {
 	      default: // LINEAR, LITERAL, or MANDATORY:
 		cond = true;
@@ -779,7 +863,9 @@ public class PrettyWriter extends java.io.Writer
 		    cond = true;
 		    break;
 		  }
-		int end = queueInts[next+QUEUED_OP_SECTION_START_SECTION_END];
+		/// ... fall through to ...
+	      case NEWLINE_SPACE:
+		int end = queueInts[next+QITEM_SECTION_START_SECTION_END];
 		if (end == 0)
 		  end = -1;
 		else
@@ -810,64 +896,71 @@ public class PrettyWriter extends java.io.Writer
 		  }
 	      }
 	    break;
-	  case QUEUED_OP_INDENTATION_TYPE:
+	  case QITEM_INDENTATION_TYPE:
 	    if (! isMisering())
 	      {
-		int kind = queueInts[next+QUEUED_OP_INDENTATION_KIND];
-		int indent = queueInts[next+QUEUED_OP_INDENTATION_AMOUNT];
-		if (kind == QUEUED_OP_INDENTATION_BLOCK)
+		int kind = queueInts[next+QITEM_INDENTATION_KIND];
+		int indent = queueInts[next+QITEM_INDENTATION_AMOUNT];
+		if (kind == QITEM_INDENTATION_BLOCK)
 		  indent += getStartColumn();
 		else
-		  indent += posnColumn(queueInts[next+QUEUED_OP_POSN]);
+		  indent += posnColumn(queueInts[next+QITEM_POSN]);
+		//log("setIndent from "+next+": "+queueInts[next+QITEM_INDENTATION_AMOUNT]+" column:"+indent);
 		setIndentation(indent);
 	      }
 	    break;
-	  case QUEUED_OP_BLOCK_START_TYPE:
-	    int end = queueInts[next + QUEUED_OP_SECTION_START_SECTION_END];
+	  case QITEM_BLOCK_START_TYPE:
+	    int start = next;
+	    int end = queueInts[next + QITEM_SECTION_START_SECTION_END];
 	    // Convert relative offset to absolute index:
 	    end = end > 0 ? (end + next) % queueInts.length : -1;
 	    int fits = fitsOnLine (end, forceNewlines);
+	    //log("block-start @"+next+" end:"+end+" force:"+forceNewlines+" fits:"+fits);
 	    if (fits > 0)
 	      {
 		// Just nuke the whole logical block and make it look
 		// like one nice long literal.
-		int endr = queueInts[next + QUEUED_OP_BLOCK_START_BLOCK_END];
+		int endr = queueInts[next + QITEM_BLOCK_START_BLOCK_END];
 		// make absolute:
 		next = (endr + next) % queueInts.length;
 		expandTabs(next);
 		queueTail = next;
 		queueSize -= endr;
+		//log("remove block -> next:"+next+" endr:"+endr+" qSize:"+queueSize);
 	      }
 	    else if (fits < 0)
 	      {
-		String prefix = queueStrings[next + QUEUED_OP_BLOCK_PREFIX];
-		String suffix = queueStrings[next + QUEUED_OP_BLOCK_START_SUFFIX];
-		reallyStartLogicalBlock (posnColumn(queueInts[next + QUEUED_OP_POSN]),
+		String prefix = queueStrings[next + QITEM_BLOCK_START_PREFIX];
+		String suffix = queueStrings[next + QITEM_BLOCK_START_SUFFIX];
+		//log("reallyStartLogicalBlock: "+blockDepth+" at:"+next);
+		reallyStartLogicalBlock (posnColumn(queueInts[next + QITEM_POSN]),
 					 prefix, suffix);
-					 
 	      }
 	    else // Don't know.
 	      break loop;
+	    if (currentBlock == start)
+	      currentBlock = -1;
 	    break;
-	  case QUEUED_OP_BLOCK_END_TYPE:
+	  case QITEM_BLOCK_END_TYPE:
+	    //log("reallyEndLogicalBlock: "+blockDepth+" at:"+next);
 	    reallyEndLogicalBlock();
 	    break;
-	  case QUEUED_OP_TAB_TYPE:
+	  case QITEM_TAB_TYPE:
 	    expandTabs(next);
 	    break;
 	  }
 	int size = getQueueSize(queueTail);
 	queueSize -= size;
+	//log("maybeWrite size: "+size+" ->"+queueSize);
 	queueTail = next + size;
       }
     return outputAnything;
   }
 
-  int miserWidth;
   protected int getMiserWidth () // DONE
   {
     // CommonLisp:  Use *print-miser-width*.
-    return 40;
+    return miserWidth;
   }
 
   boolean isMisering() // DONE
@@ -899,7 +992,7 @@ public class PrettyWriter extends java.io.Writer
 	available -= getSuffixLength();
       }
     if (sectionEnd >= 0)
-      return posnColumn(queueInts[sectionEnd + QUEUED_OP_POSN]) <= available ? 1 : -1;
+      return posnColumn(queueInts[sectionEnd + QITEM_POSN]) <= available ? 1 : -1;
     if (forceNewlines)
       return -1;
     if (indexColumn(bufferFillPointer) > available)
@@ -918,9 +1011,9 @@ public class PrettyWriter extends java.io.Writer
   void outputLine (int newline)  throws IOException
   {
     char[] buffer = this.buffer;
-    int kind = queueInts[newline + QUEUED_OP_NEWLINE_KIND];
+    int kind = queueInts[newline + QITEM_NEWLINE_KIND];
     boolean isLiteral = kind == NEWLINE_LITERAL;
-    int amountToConsume = posnIndex(queueInts[newline + QUEUED_OP_POSN]);
+    int amountToConsume = posnIndex(queueInts[newline + QITEM_POSN]);
     int amountToPrint;
     if (isLiteral)
       amountToPrint = amountToConsume;
@@ -943,6 +1036,7 @@ public class PrettyWriter extends java.io.Writer
       }
     out.write(buffer, 0, amountToPrint);
     int lineNumber = this.lineNumber;
+    //log("outputLine#"+lineNumber+": \""+new String(buffer, 0, amountToPrint)+"\" curBlock:"+currentBlock);
     lineNumber++;
     if (! printReadably())
       {
@@ -992,7 +1086,7 @@ public class PrettyWriter extends java.io.Writer
   {
     int fillPtr = bufferFillPointer;
     int tail = queueTail;
-    while (queueSize > 0 && getQueueType(tail) == QUEUED_OP_NOP_TYPE)
+    while (queueSize > 0 && getQueueType(tail) == QITEM_NOP_TYPE)
       {
 	int size = getQueueSize(tail);
 	queueSize -= size;
@@ -1001,7 +1095,7 @@ public class PrettyWriter extends java.io.Writer
 	  tail = 0;
 	queueTail = tail;
       }
-    int count = queueSize > 0 ? posnIndex (queueInts[tail + QUEUED_OP_POSN])
+    int count = queueSize > 0 ? posnIndex (queueInts[tail + QITEM_POSN])
       : fillPtr;
     int newFillPtr = fillPtr - count;
     if (count <= 0)
@@ -1009,6 +1103,7 @@ public class PrettyWriter extends java.io.Writer
     try
       {
 	out.write(buffer, 0, count);
+	//log("outputPartial: \""+new String(buffer, 0, count)+'\"');
       }
     catch (IOException ex)
       {
@@ -1087,80 +1182,110 @@ public class PrettyWriter extends java.io.Writer
     pendingBlocksCount = 0;
   }
 
+  /*
+  public static PrintWriter log;
+  static {
+    try { log = new PrintWriter(new FileOutputStream("/tmp/pplog")); }
+    catch (Throwable ex) { ex.printStackTrace(); }
+  }
+  void log(String str)
+  {
+    log.println(str);
+    log.flush();
+  }
   void dumpQueue()
   {
-    PrintWriter out = new PrintWriter(System.err);
-    out.println("Queue tail:"+queueTail+" size:"+queueSize);
-    dumpQueue(queueTail, queueSize, out);
-    out.flush();
+    log.println("Queue tail:"+queueTail+" size:"+queueSize
+		+" length:"+queueInts.length+" startCol:"+bufferStartColumn);
+    dumpQueue(queueTail, queueSize, log);
   }
 
   void dumpQueue(int start, int todo, PrintWriter out)
   {
+    int bufIndex = 0;
     while (todo > 0)
       {
 	if (start == queueInts.length)
 	  start = 0;
+	if (start < 0 || start >= queueInts.length)
+	  {
+	    out.print('@');	out.print(start);  out.print(": ");
+	    out.print("out of bounds - queueInts length is ");
+	    out.println(queueInts.length);
+	    break;
+	  }
 	int type = getQueueType(start);
 	int size = getQueueSize(start);
+	if (type != QITEM_NOP_TYPE)
+	  {
+	    int newIndex = posnIndex(queueInts[start+QITEM_POSN]);
+	    int count = newIndex - bufIndex;
+	    if (count > 0)
+	      {
+		out.print(count); out.print(" chars: \"");
+		out.write(buffer, bufIndex, count);
+		out.println('\"');
+		bufIndex = newIndex;
+	      }
+	  }
 	out.print('@');	out.print(start);  out.print(": ");
 	out.print("type:");  out.print(type);
 	switch (type)
 	  {
-	  case QUEUED_OP_NEWLINE_TYPE:
+	  case QITEM_NEWLINE_TYPE:
 	    out.print("(newline)");  break;
-	  case QUEUED_OP_INDENTATION_TYPE:
+	  case QITEM_INDENTATION_TYPE:
 	    out.print("(indentation)");  break;
-	  case QUEUED_OP_BLOCK_START_TYPE:
+	  case QITEM_BLOCK_START_TYPE:
 	    out.print("(block-start)");  break;
-	  case QUEUED_OP_BLOCK_END_TYPE:
+	  case QITEM_BLOCK_END_TYPE:
 	    out.print("(block-end)");  break;
-	  case QUEUED_OP_TAB_TYPE:  out.print("(tab)");
+	  case QITEM_TAB_TYPE:  out.print("(tab)");
 	    break;
-	  case QUEUED_OP_NOP_TYPE:
+	  case QITEM_NOP_TYPE:
 	    out.print("(nop)");  break;
 	  }
 	out.print(" size:");  out.print(size);
-	out.print(";  @");  out.print(start+QUEUED_OP_POSN);
-	if (type != QUEUED_OP_NOP_TYPE)
+	out.print(";  @");  out.print(start+QITEM_POSN);
+	if (type != QITEM_NOP_TYPE)
 	  {
 	    out.print(": posn:");
-	    int posn = queueInts[start+QUEUED_OP_POSN];
+	    int posn = queueInts[start+QITEM_POSN];
 	    out.print(posn);
 	    out.print(" index:");
 	    out.println(posnIndex(posn));
 	  }
-	if (type == QUEUED_OP_NEWLINE_TYPE
-	    || type == QUEUED_OP_BLOCK_START_TYPE)
+	if (type == QITEM_NEWLINE_TYPE
+	    || type == QITEM_BLOCK_START_TYPE)
 	    
 	  {
-	    out.print('@');  out.print(start+QUEUED_OP_SECTION_START_DEPTH);
+	    out.print('@');  out.print(start+QITEM_SECTION_START_DEPTH);
 	    out.print(": - depth:");
-	    out.print(queueInts[start+QUEUED_OP_SECTION_START_DEPTH]);
+	    out.print(queueInts[start+QITEM_SECTION_START_DEPTH]);
 	    out.print(";  @");
-	    out.print(start+QUEUED_OP_SECTION_START_SECTION_END);
+	    out.print(start+QITEM_SECTION_START_SECTION_END);
 	    out.print(": section-end:");
-	    out.println(queueInts[start+QUEUED_OP_SECTION_START_SECTION_END]);
+	    out.println(queueInts[start+QITEM_SECTION_START_SECTION_END]);
 	  }
 	switch (type)
 	  {
-	  case QUEUED_OP_BLOCK_START_TYPE:
-	    printQueueWord(start, QUEUED_OP_BLOCK_START_BLOCK_END, "block-end", out);
-	    String prefix = queueStrings[start+QUEUED_OP_BLOCK_PREFIX];
-	    printQueueStringWord(start, QUEUED_OP_BLOCK_PREFIX, "prefix", out);
-	    printQueueStringWord(start, QUEUED_OP_BLOCK_START_SUFFIX, "suffix", out);
+	  case QITEM_BLOCK_START_TYPE:
+	    printQueueWord(start, QITEM_BLOCK_START_BLOCK_END, "block-end", out);
+	    printQueueStringWord(start, QITEM_BLOCK_START_PREFIX, "prefix", out);
+	    printQueueStringWord(start, QITEM_BLOCK_START_SUFFIX, "suffix", out);
 	    break;
-	  case QUEUED_OP_NEWLINE_TYPE:
+	  case QITEM_NEWLINE_TYPE:
 	    out.print('@');
-	    out.print(start+QUEUED_OP_NEWLINE_KIND);
+	    out.print(start+QITEM_NEWLINE_KIND);
 	    out.print(": - kind: ");
-	    int kind = queueInts[start+QUEUED_OP_NEWLINE_KIND];
+	    int kind = queueInts[start+QITEM_NEWLINE_KIND];
 	    String skind = "???";
 	    switch (kind)
 	      {
 	      case NEWLINE_LINEAR:    skind = "linear";    break;
 	      case NEWLINE_LITERAL:   skind = "literal";   break;
 	      case NEWLINE_FILL:      skind = "fill";      break;
+	      case NEWLINE_SPACE:     skind = "space";      break;
 	      case NEWLINE_MISER:     skind = "miser";     break;
 	      case NEWLINE_MANDATORY: skind = "mandatory"; break;
 	      }
@@ -1176,6 +1301,13 @@ public class PrettyWriter extends java.io.Writer
 	todo -= size;
 	start += size;
       }
+    int count = bufferFillPointer - bufIndex;
+    if (count > 0)
+      {
+	out.print(count); out.print(" chars: \"");
+	out.write(buffer, bufIndex, count);
+	out.println('\"');
+      }
   }
 
   private void printQueueWord(int start, int offset,
@@ -1209,4 +1341,41 @@ public class PrettyWriter extends java.io.Writer
 	out.println(str.length());
       }
   }
+
+  void check (String where)
+  {
+    String msg = null;
+    if (currentBlock != -1
+	&& ! (currentBlock < queueInts.length
+	      && currentBlock >= queueTail
+	      ? currentBlock < queueTail + queueSize
+	      : currentBlock < queueTail + queueSize - queueInts.length))
+      msg = ("currentBlock ("+currentBlock
+	     +") >= queue length ("+queueInts.length+")");
+    if (msg != null)
+      {
+	if (where != null)
+	  msg = "in "+where+": "+msg;
+	log.println(msg);
+	dumpQueue(); 
+	log.flush();
+	throw new Error(msg);
+      }
+  }
+
+  String itemKindString (int kind)
+  {
+    switch (kind)
+      {
+      case QITEM_NOP_TYPE:  return "nop";
+      case QITEM_NEWLINE_TYPE:  return "newline";
+      case QITEM_INDENTATION_TYPE:  return "indentation";
+      case QITEM_BLOCK_START_TYPE:  return "block-start";
+      case QITEM_BLOCK_END_TYPE:  return "block-end";
+      case QITEM_TAB_TYPE:  return "tab";
+      default: return "("+kind+" - unknown)";
+      }
+  }
+  String enqueueExtraLog = "";
+  */
 }
