@@ -11,17 +11,24 @@ import gnu.lists.*;
 
 abstract public class Syntax implements Printable, Named
 {
-  String name;
+  Object name;
 
-  public String getName() { return name; }
+  public final String getName()
+  {
+    return name == null ? null
+      : name instanceof Symbol ? ((Symbol) name).getName()
+      : name.toString();
+  }
+  public Object getSymbol() { return name; }
 
+  public void setName (Object name) { this.name = name; }
   public void setName (String name) { this.name = name; }
 
   public Syntax ()
   {
   }
 
-  public Syntax (String name)
+  public Syntax (Object name)
   {
     setName(name);
   }
@@ -39,14 +46,30 @@ abstract public class Syntax implements Printable, Named
     throw new InternalError("rewrite method not defined");
   }
 
+  public Expression rewriteForm (Object form, Translator tr)
+  {
+    if (form instanceof Pair)
+      return rewriteForm((Pair) form, tr);
+    else
+      return tr.syntaxError("non-list form for "+this);
+  }
+
   public Expression rewriteForm (Pair form, Translator tr)
   {
     return rewrite(form.cdr, tr);
   }
 
+  public void scanForm (Pair st, ScopeExp defs, Translator tr)
+  {
+    boolean ok = scanForDefinitions(st, tr.formStack, defs, tr);
+    if (! ok)
+      tr.formStack.add(new ErrorExp("syntax error"));
+  }
+
   /** Check if a statement is a definition, for initial pass.
+   * Semi-deprecated - should convert calls to use scanForm.
    * @param st the statement to check
-   * @param form where to append the (possibly-modified) statement
+   * @param forms where to append the (possibly-modified) statement
    * @param defs where to add Declarations for found definitions
    * @param tr the compilation state
    * @return true on success
