@@ -14,12 +14,13 @@
   (and (instance? x <number>) (not (invoke (as <number> x) 'isExact))))
 
 (define (zero? (x :: <number>)) :: <boolean> 
-  ((primitive-virtual-method <number> "isZero" <boolean> ())
-   x))
+  (invoke x 'isZero))
+
+(define (positive? (x :: <real>)) :: <boolean>
+  (> (invoke x 'sign) 0))
 
 (define (negative? (x :: <real>)) :: <boolean> 
-  ((primitive-virtual-method <real> "isNegative" <boolean> ())
-   x))
+  (invoke x 'isNegative))
 
 (define (odd? (x :: <integer>)) :: <boolean> 
   ((primitive-virtual-method <integer> "isOdd" <boolean> ())
@@ -31,11 +32,26 @@
 (define (abs (x :: <number>)) :: <number>
   (invoke x 'abs))
 
-(define (quotient (x :: <integer>) (y :: <integer>)) :: <integer>
-  (invoke-static <integer> 'quotient x y))
+(define (quotient (x :: <real>) (y :: <real>)) :: <real>
+  (if (and (instance? x <integer>) (instance? y <integer>))
+      (invoke-static <integer> 'quotient x y)
+      (invoke (/ x y) 'toInt (static-field <number> 'TRUNCATE))))
 
-(define (remainder (x :: <integer>) (y :: <integer>)) :: <integer>
-  (invoke-static <integer> 'remainder x y))
+(define (remainder (x :: <real>) (y :: <real>)) :: <real> 
+  (if (and (instance? x <integer>) (instance? y <integer>))
+      (invoke-static <integer> 'remainder x y)
+      (if (zero? y)
+	  (if (exact? y) x (exact->inexact x))
+	  (- x (* (invoke (/ x y) 'toInt (static-field <number> 'TRUNCATE))
+		  y)))))
+
+(define (modulo (x :: <real>) (y :: <real>)) :: <real>
+  (if (and (instance? x <integer>) (instance? y <integer>))
+      (invoke-static <integer> 'modulo x y)
+      (if (zero? y)
+	  (if (exact? y) x (exact->inexact x))
+	  (- x (* (invoke (/ x y) 'toInt (static-field <number> 'FLOOR))
+		  y)))))
 
 (define (numerator (x :: <rational>)) :: <integer>
   (invoke x 'numerator))
@@ -55,6 +71,11 @@
 (define (round (x :: <real>)) :: <real>
   (invoke x 'toInt (static-field <number> 'ROUND)))
 
+(define (rationalize (x :: <real>) (y :: <real>)) :: <real>
+  (gnu.math.RatNum:rationalize
+   (as <real> (invoke x 'sub y))
+   (as <real> (invoke x 'add y))))
+
 (define (exp (x :: <complex>)) :: <complex>
   (invoke x 'exp))
 
@@ -77,6 +98,11 @@
 (define (acos (x :: <double>)) :: <double>
   (invoke-static <java.lang.Math> 'acos x))
 
+(define (sqrt (num :: <quantity>)) :: <quantity>
+  (gnu.math.Quantity:make
+   (invoke (invoke num 'number) 'sqrt)
+   (invoke (invoke num 'unit) 'sqrt)))
+
 (define (make-rectangular (x :: <real>) (y :: <real>)) :: <complex>
   (invoke-static <complex> 'make x y))
 
@@ -93,7 +119,17 @@
   (invoke x 'abs))
 
 (define (angle (x :: <complex>)) :: <real>
-  (invoke (invoke x 'number) 'angle))
+  (invoke x 'angle))
+
+(define (exact->inexact (num :: <number>)) :: <number>
+  (if (invoke num 'isExact)
+      (make <gnu.math.DFloNum> (invoke (as <real> num) 'doubleValue))
+      num))
+
+(define (inexact->exact (num :: <number>)) :: <number>
+  (if (instance? num <real>)
+      (invoke (as <real> num) 'toExact)
+      num))
 
 (define (arithmetic-shift (value :: <integer>) (amount :: <int>)) :: <integer>
   (invoke-static <integer> 'shift value amount))
