@@ -1,6 +1,5 @@
 package kawa;
-import kawa.lang.*;
-import kawa.standard.*;
+import kawa.lang.CompileFile;
 import java.io.*;
 import gnu.mapping.*;
 import gnu.expr.*;
@@ -16,23 +15,24 @@ public class repl extends Procedure0or1
   public static String compilationTopname = null;
   public static String compilationPrefix = null;
 
-  Interpreter interp;
+  Language language;
+  static Language previousLanguage;
 
-  public repl(Interpreter interp)
+  public repl(Language language)
   {
-    this.interp = interp;
+    this.language = language;
   }
 
   public Object apply0 ()
   {
-    Shell.run(interp, Environment.getCurrent());
-    return Scheme.voidObject;
+    Shell.run(language, Environment.getCurrent());
+    return Values.empty;
   }
 
   public Object apply1(Object env)
   {
-    Shell.run(interp, (Environment) env);
-    return Scheme.voidObject;
+    Shell.run(language, (Environment) env);
+    return Values.empty;
   }
 
   static void bad_option (String str)
@@ -42,29 +42,42 @@ public class repl extends Procedure0or1
     System.exit (-1);
   }
 
+  public static void printOption(PrintStream out, String option, String doc)
+  {
+    out.print(" ");
+    out.print(option);
+
+    int len = option.length() + 1;
+    for (int i = 0; i < 30 - len; ++i)
+      out.print(" ");
+    out.print(" ");
+    out.println(doc);
+  }
+  
   public static void printOptions(PrintStream out)
   {
     out.println("Usage: [java kawa.repl | kawa] [options ...]");
     out.println();
     out.println(" Generic options:");
-    out.println(" --help                    Show help about options");
-    out.println(" --author                  Show author information");
-    out.println(" --version                 Show version information");
+    printOption(out, "--help", "Show help about options");
+    printOption(out, "--author", "Show author information");
+    printOption(out, "--version", "Show version information");
     out.println();
     out.println(" Options");
-    out.println(" -e <expr>                 Evaluate expression <expr>");
-    out.println(" -c <expr>                 Same as -e, but make sure ~/.kawarc.scm is run first");
-    out.println(" -f <filename>             File to interpret");
-    out.println(" -s | --                   Start reading commands interactively from console");
-    out.println(" -w                        Launch the interpreter in a GUI window");
-    out.println(" --server <port>           Start a server accepting telnet connections on <port>");
-    out.println(" --debug-dump-zip          Compiled interactive expressions to a zip archive");
-    out.println(" --debug-print-expr        Print generated internal expressions");
-    out.println(" --debug-print-final-expr  Print expression after any optimizations");
-    out.println(" --[no-]full-tailcalls     (Don't) use full tail-calls");
-    out.println(" -C <filename> ...         Compile named files to Java class files");
-    out.println(" --<language>              Select source language, one of:");
-    String[][] languages = Interpreter.getLanguages();
+    printOption(out, "-e <expr>", "Evaluate expression <expr>");
+    printOption(out, "-c <expr>", "Same as -e, but make sure ~/.kawarc.scm is run first");
+    printOption(out, "-f <filename>", "File to interpret");
+    printOption(out, "-s| --", "Start reading commands interactively from console");
+    printOption(out, "-w", "Launch the interpreter in a GUI window");
+    printOption(out, "--server <port>", "Start a server accepting telnet connections on <port>");
+    printOption(out, "--debug-dump-zip", "Compiled interactive expressions to a zip archive");
+    printOption(out, "--debug-print-expr", "Print generated internal expressions");
+    printOption(out, "--debug-print-final-expr", "Print expression after any optimizations");
+    printOption(out,"--[no-]full-tailcalls", "(Don't) use full tail-calls");
+    printOption(out, "-C <filename> ...", "Compile named files to Java class files");
+    printOption(out, "--output-format <format>", "Use <format> when printing top-level output");
+    printOption(out,"--<language>", "Select source language, one of:");
+    String[][] languages = Language.getLanguages();
     for (int i = 0; i < languages.length; i++)
       {
 	out.print("   ");
@@ -78,18 +91,57 @@ public class repl extends Procedure0or1
 	out.println();
       }
     out.println(" Compilation options, must be specified before -C");
-    out.println(" -d <dirname>              Directory to place .class files in");
-    out.println(" -P <prefix>               Prefix to prepand to class names");
-    out.println(" -T <topname>              name to give ot top-level class");
+    printOption(out, "-d <dirname>", "Directory to place .class files in");
+    printOption(out, "-P <prefix>", "Prefix to prepand to class names");
+    printOption(out, "-T <topname>", "name to give to top-level class");
     
-    out.println(" --main                    Generate an application, with a main method");
-    out.println(" --applet                  Generate an applet");
-    out.println(" --servlet                 Generate a servlet");
-    out.println(" --module-static           Top-leval definitions are by default static");
+    printOption(out, "--main", "Generate an application, with a main method");
+    printOption(out, "--applet", "Generate an applet");
+    printOption(out, "--servlet", "Generate a servlet");
+    printOption(out, "--module-static", "Top-leval definitions are by default static");
+
+    Vector keys = Compilation.options.keys();
+    for (int i = 0; i < keys.size(); ++i)
+      {
+        String name = (String) keys.get(i);
+        printOption(out, "--" + name, Compilation.options.getDoc(name));
+      }
+        
     out.println();
     out.println("For more information go to:  http://www.gnu.org/software/kawa/");
   }
 
+  /** Number of times exitDecrement calls before we exit. */
+  private static int exitCounter;
+  /** See exitDecrement. */
+  public static synchronized void exitIncrement()
+  {
+    if (exitCounter == 0)
+      exitCounter++;
+    exitCounter++;
+  }
+
+  /** Work around an AWT bug, where AWT threads are non-daemon.
+   * Thus if you start up AWT, the JVM will wait for the AWT to finish,
+   * even if there are no other non-daemon threads.
+   * So call exitIncrement() each time a Freme is created,
+   * and call exitDecrement() a Frame is closed. */
+  public static synchronized void exitDecrement()
+  {
+    int counter = exitCounter;
+    if (counter > 0)
+      {
+	counter--;
+	if (counter == 0)
+	  {
+	    System.exit(0);
+	  }
+	else
+	  exitCounter = counter;
+      }
+  }
+
+  public static String[] commandLineArgArray;
   public static FVector commandLineArguments;
 
   public static String homeDirectory;
@@ -112,8 +164,8 @@ public class repl extends Procedure0or1
 	    initFile = new File(homeDirectory, kawarc_name);
 	  }
 	else
-	  scmHomeDirectory = Scheme.falseObject;
-	Environment.define_global("home-directory", scmHomeDirectory);
+	  scmHomeDirectory = Boolean.FALSE;
+	Environment.getCurrent().put("home-directory", scmHomeDirectory);
 	if (initFile != null && initFile.exists())
 	  Shell.runFile(initFile.getPath());
       }
@@ -121,44 +173,50 @@ public class repl extends Procedure0or1
 
   public static void setArgs (String[] args, int arg_start)
   {
-    Object[] array = new Object[args.length - arg_start];
-    for (int i = arg_start;  i < args.length;  i++)
-      array[i - arg_start] = new FString (args[i]);
+    int nargs = args.length - arg_start;
+    Object[] array = new Object[nargs];
+    if (arg_start == 0)
+      commandLineArgArray = args;
+    else
+      {
+	String[] strings = new String[nargs];
+	for (int i = nargs;  --i >= 0; )
+	  strings[i] = args[i+arg_start];
+	commandLineArgArray = strings;
+      }
+    for (int i = nargs;  --i >= 0; )
+      array[i] = new FString (args[i + arg_start]);
     commandLineArguments = new FVector (array);  // FIXME scsh has list
     // FIXME scsh also has command-line proc
-    Environment.define_global ("command-line-arguments",
-			       commandLineArguments);
+    Environment.getCurrent().put("command-line-arguments",
+				 commandLineArguments);
   }
 
-  public static Interpreter getInterpreterFromFilenameExtension(String name)
+  public static void getLanguageFromFilenameExtension(String name)
   {
-    if (Interpreter.defaultInterpreter == null)
+    if (previousLanguage == null)
       {
-	Interpreter interp
-	  = Interpreter.getInstanceFromFilenameExtension(name);
-	if (interp != null)
+	previousLanguage = Language.getInstanceFromFilenameExtension(name);
+	if (previousLanguage != null)
 	  {
-	    Interpreter.defaultInterpreter = interp;
-	    Environment.setCurrent(interp.getEnvironment());
-	    return interp;
+	    Language.setDefaults(previousLanguage);
+	    return;
 	  }
       }
-    return getInterpreter();
+    getLanguage();
   }
 
-  public static Interpreter getInterpreter()
+  public static void getLanguage()
   {
-    if (Interpreter.defaultInterpreter == null)
+    if (previousLanguage == null)
       {
-	Interpreter.defaultInterpreter = Interpreter.getInstance(null);
-	Environment.setCurrent(Interpreter.defaultInterpreter.getEnvironment());
+	previousLanguage = Language.getInstance(null);
+	Language.setDefaults(previousLanguage);
       }
-    return Interpreter.defaultInterpreter;
   }
 
   static boolean shutdownRegistered
     = gnu.text.WriterManager.instance.registerShutdownHook();
-  static boolean windowStarted = false;
 
   public static int processArgs(String[] args, int iArg, int maxArg)
   {
@@ -171,12 +229,12 @@ public class repl extends Procedure0or1
 	    iArg++;
 	    if (iArg == maxArg)
 	      bad_option (arg);
-	    getInterpreter();
+	    getLanguage();
 	    setArgs (args, iArg+1);
 	    if (arg.equals ("-c"))
 	      checkInitFile();
-	    Interpreter interp = Interpreter.defaultInterpreter;
-	    Shell.runString(args[iArg], interp, interp.getEnvironment());
+	    Language language = Language.getDefaultLanguage();
+	    Shell.runString(args[iArg], language, Environment.getCurrent());
 	    something_done = true;
 	  }
 	else if (arg.equals ("-f"))
@@ -185,7 +243,7 @@ public class repl extends Procedure0or1
 	    if (iArg == maxArg)
 	      bad_option (arg);
 	    String filename = args[iArg];
-	    getInterpreterFromFilenameExtension(filename);
+	    getLanguageFromFilenameExtension(filename);
 	    setArgs (args, iArg+1);
 	    checkInitFile();
 	    Shell.runFile (filename);
@@ -259,7 +317,7 @@ public class repl extends Procedure0or1
 			  }
 		      }
 		  }
-		getInterpreterFromFilenameExtension(filename);
+		getLanguageFromFilenameExtension(filename);
 		freader = InPort.openFile(fstream, filename);
 		// FIXME adjust line number
 		setArgs(args, iArg+1);
@@ -286,16 +344,16 @@ public class repl extends Procedure0or1
 	else if (arg.equals ("-s") || arg.equals ("--"))
 	  {
 	    iArg++;
-	    getInterpreter();
+	    getLanguage();
 	    setArgs (args, iArg);
 	    checkInitFile();
-	    Shell.run(Interpreter.defaultInterpreter, Environment.getCurrent());
+	    Shell.run(Language.getDefaultLanguage(), Environment.getCurrent());
 	    return -1;
 	  }
 	else if (arg.equals ("-w"))
 	  {
 	    iArg++;
-	    getInterpreter();
+	    getLanguage();
 	    setArgs (args, iArg);
 	    checkInitFile();
 	    // Do this instead of just new GuiConsole in case we have
@@ -303,7 +361,6 @@ public class repl extends Procedure0or1
 	    try
 	      {
 		Class.forName("kawa.GuiConsole").newInstance();
-		windowStarted = true;
 	      }
 	    catch (Exception ex)
 	      {
@@ -341,7 +398,7 @@ public class repl extends Procedure0or1
 	    for ( ; iArg < maxArg;  iArg++)
 	      {
 		arg = args[iArg];
-		getInterpreterFromFilenameExtension(arg);
+		getLanguageFromFilenameExtension(arg);
 		try
 		  {
 		    System.err.println("(compiling "+arg+")");
@@ -412,7 +469,7 @@ public class repl extends Procedure0or1
 	  }
 	else if (arg.equals("--server"))
 	  {
-	    getInterpreter();
+	    getLanguage();
 	    ++iArg;
 	    if (iArg == maxArg)
 	      bad_option (arg);
@@ -444,7 +501,7 @@ public class repl extends Procedure0or1
 		    System.err.println("got connection from "
 				       +client.getInetAddress()
 				       +" port:"+client.getPort());
-		    serveTelnet(Interpreter.defaultInterpreter, client);
+		    TelnetRepl.serve(Language.getDefaultLanguage(), client);
 		  }
 	      }
 	    catch (java.io.IOException ex)
@@ -480,23 +537,38 @@ public class repl extends Procedure0or1
 	  {
 	    gnu.expr.Compilation.moduleStatic = 1;
 	  }
+        else if (arg.equals("--module-static-run"))
+          {
+            gnu.expr.Compilation.moduleStatic = 2;
+          }
 	else if (arg.equals("--fewer-classes"))
 	  {
 	    gnu.expr.Compilation.fewerClasses = true;
 	  }
+	else if (arg.equals("--no-inline")
+		 || arg.equals("--inline=none"))
+	  {
+	    gnu.expr.Compilation.inlineOk = false;
+	  }
+	else if (arg.equals("--inline"))
+	  {
+	    gnu.expr.Compilation.inlineOk = true;
+	  }
 	else if (arg.equals("--cps"))
 	  {
-	    gnu.expr.Compilation.fewerClasses = true;
-	    gnu.expr.Compilation.usingTailCalls = true;
-	    gnu.expr.Compilation.usingCPStyle = true;
+	    Compilation.fewerClasses = true;
+	    Compilation.defaultCallConvention
+	      = Compilation.CALL_WITH_CONTINUATIONS;
 	  }
 	else if (arg.equals("--full-tailcalls"))
 	  {
-	    gnu.expr.Compilation.usingTailCalls = true;
+	    Compilation.defaultCallConvention
+	      = Compilation.CALL_WITH_TAILCALLS;
 	  }
 	else if (arg.equals("--no-full-tailcalls"))
 	  {
-	    gnu.expr.Compilation.usingTailCalls = false;
+	    Compilation.defaultCallConvention
+	      = Compilation.CALL_WITH_RETURN;
 	  }
 	else if (arg.equals("--help"))
 	  {
@@ -513,31 +585,142 @@ public class repl extends Procedure0or1
 	    System.out.print("Kawa ");
 	    System.out.print(Version.getVersion());
 	    System.out.println();
-	    System.out.println("Copyright (C) 2001 Per Bothner");
+	    System.out.println("Copyright (C) 2005 Per Bothner");
 	    something_done = true;
 	  }
 	else if (arg.length () > 0 && arg.charAt(0) == '-')
 	  { // Check if arg is a known language name.
-	    Interpreter previous = Interpreter.defaultInterpreter;
 	    String name = arg;
 	    if (name.length() > 2 && name.charAt(0) == '-')
 	      name = name.substring(name.charAt(1) == '-' ? 2 :1);
-	    Interpreter interpreter = Interpreter.getInstance(name);
-	    if (interpreter == null)
-	      bad_option(arg);
+	    Language lang = Language.getInstance(name);
+	    if (lang != null)
+	      {
+		if (previousLanguage == null)
+		  Language.setDefaults(lang);
+                else
+                  Language.setDefaultLanguage(lang);
+		previousLanguage = lang;
+	      }
 	    else
 	      {
-		Interpreter.defaultInterpreter = interpreter;
-		if (previous == null)
-		  Environment.setCurrent(interpreter.getEnvironment());
+		// See if arg is a valid Compilation option, and if so set it.
+		int eq = name.indexOf("=");
+		String opt_value;
+		if (eq < 0)
+		  opt_value = null;
+		else
+		  {
+		    opt_value = name.substring(eq+1);
+		    name = name.substring(0, eq);
+		  }
+
+		// Convert "--no-xxx" to "--xxx=no":
+		boolean startsWithNo
+		  = name.startsWith("no-") && name.length() > 3;
+		if (opt_value == null && startsWithNo)
+		  {
+		    opt_value = "no";
+		    name = name.substring(3);
+		  }
+
+		String msg = Compilation.options.set(name, opt_value);
+		if (msg != null)
+		  {
+		    // It wasn't a valid Complation option.
+		    if (startsWithNo && msg == gnu.text.Options.UNKNOWN)
+		      msg = "both '--no-' prefix and '="+
+			opt_value+"' specified";
+		    if (msg == gnu.text.Options.UNKNOWN)
+		      {
+			bad_option(arg);
+		      }
+		    else
+		      {
+			System.err.println ("kawa: bad option '"
+					    + arg + "': " + msg);
+			System.exit (-1);
+		      }
+		  }
 	      }
 	  }
 	else
-	  return iArg;
+          {
+            int eq = arg.indexOf('=');
+            if (eq <= 0)
+              return iArg;
+            String key = arg.substring(0, eq);
+            String value = arg.substring(eq+1);
+            String uri, local;
+            for (int i = 0; ; i++)
+              {
+                String[] propertyField = propertyFields[i];
+                if (propertyField == null)
+                  break;
+                if (key.equals(propertyField[0]))
+                  {
+                    String cname = propertyField[1];
+                    String fname = propertyField[2];
+                    try
+                      {
+                        Class clas = Class.forName(cname);
+                        ThreadLocation loc = (ThreadLocation)
+                          clas.getDeclaredField(fname).get(null);
+                        loc.setGlobal(value);
+                        continue;
+                      }
+                    catch (Throwable ex)
+                      {
+                        System.err.println("error setting property " + key
+                                           +" field "+cname+'.'+fname+": "+ex);
+                        System.exit(-1);
+                      }
+                  }
+              }
+            if (key.charAt(0) == '{')
+              {
+                int rbrace = key.lastIndexOf('}');
+                if (rbrace <= 0)
+                  {
+                    System.err.println("missing '}' in property name '"+key+"'");
+                    System.exit (-1);
+                  }
+                uri = key.substring(1, rbrace);
+                local = key.substring(rbrace+1);
+              }
+            else
+              {
+                uri = "";
+                local = key;
+              }
+            Symbol symbol = Symbol.make(uri, local);
+            // Run Language's static initializer.
+            Language.getDefaultLanguage();
+            Environment current = Environment.getCurrent();
+            current.put(symbol, null, value);
+          }
       }
     return something_done ? -1 : iArg;
   }
 
+  /** A list of standard command-line fluid names to map to static fields.
+   * For each entry:
+   * element 0 is a property name (before the '=' in the comamnd-line);
+   * element 1 is the name of a class;
+   * element 2 is the name of a static ThreadLocation field. */
+  static String[][] propertyFields =
+    {
+      { "out:doctype-system", "gnu.xml.XMLPrinter", "doctypeSystem" },
+      { "out:doctype-public", "gnu.xml.XMLPrinter", "doctypePublic" },
+      { "out:base", "gnu.kawa.functions.DisplayFormat", "outBase" },
+      { "out:radix", "gnu.kawa.functions.DisplayFormat", "outRadix" },
+      { "out:line-length", "gnu.text.PrettyWriter", "lineLengthLoc" },
+      { "out:right-margin", "gnu.text.PrettyWriter", "lineLengthLoc" },
+      { "out:miser-width", "gnu.text.PrettyWriter", "miserWidthLoc" },
+      { "out:xml-indent", "gnu.xml.XMLPrinter", "indentLoc" },
+      null
+    };
+
   public static void main(String args[])
   {
     try
@@ -548,85 +731,27 @@ public class repl extends Procedure0or1
 	if (iArg < args.length)
 	  {
 	    String filename = args[iArg];
-	    getInterpreterFromFilenameExtension(filename);
+	    getLanguageFromFilenameExtension(filename);
 	    setArgs (args, iArg+1);
 	    checkInitFile();
 	    Shell.runFile (filename);
 	  }
 	else
 	  {
-	    getInterpreter();
+	    getLanguage();
 	    setArgs (args, iArg);
 	    checkInitFile();
-	    Shell.run(Interpreter.defaultInterpreter);
+	    Shell.run(Language.getDefaultLanguage(), Environment.getCurrent());
 	  }
       }
     finally
       {
-	if (! shutdownRegistered && ! windowStarted)
+	if (! shutdownRegistered)
 	  {
 	    // Redundant if registerShutdownHook succeeded (e.g on JDK 1.3).
 	    gnu.mapping.OutPort.runCleanups();
 	  }
+	exitDecrement();
       }
    }
-
-  /** Run a Kawa repl as a telnet server.
-      @param client A client that has connected to us,
-      and that wants to use the telnet protocol to talk to a
-      Scheme read-eval-print-loop. */
-  static void serveTelnet (Interpreter interp, java.net.Socket client)
-    throws java.io.IOException
-  {
-    Telnet conn = new Telnet(client, true);
-    java.io.OutputStream sout = conn.getOutputStream();
-    java.io.InputStream sin = conn.getInputStream();
-    OutPort out = new OutPort(sout);
-    TtyInPort in = new TtyInPort(sin, "<stdin>", out);
-    /*
-    conn.request(Telnet.DO, Telnet.EOF);
-    conn.request(Telnet.DO, Telnet.NAWS);
-    conn.request(Telnet.DO, Telnet.TTYPE);
-    conn.request(Telnet.DO, Telnet.LINEMODE);
-    */
-
-    Thread thread = new Future(new SocketRepl(interp, client),
-			       interp.getEnvironment(),
-			       in, out, out);
-    thread.start();
-  }
 }
-
-class SocketRepl extends Procedure0
-{
-  // close when finished.
-  java.net.Socket socket;
-
-  Interpreter interp;
-
-  public SocketRepl(Interpreter interp, java.net.Socket socket)
-  {
-    this.interp = interp;
-    this.socket = socket;
-  }
-
-  public Object apply0 ()
-  {
-    try
-      {
-	Shell.run(interp, Environment.getCurrent());
-	return Scheme.voidObject;
-      }
-    finally
-      {
-	try
-	  {
-	    socket.close();
-	  }
-	catch (java.io.IOException ex)
-	  {
-	  }
-      }
-  }
-}
-
