@@ -1,13 +1,16 @@
 package kawa.standard;
 import kawa.lang.*;
 import gnu.lists.*;
-import gnu.mapping.*;
 import gnu.expr.*;
 import gnu.kawa.reflect.*;
 import gnu.bytecode.*;
 
 public class define_member_alias extends Syntax
 {
+  public static final define_member_alias define_member_alias
+    = new define_member_alias();
+  static { define_member_alias.setName("define-member-alias"); }
+
 
   public boolean scanForDefinitions (Pair st, java.util.Vector forms,
                                      ScopeExp defs, Translator tr)
@@ -18,10 +21,10 @@ public class define_member_alias extends Syntax
         || ! ((p = (Pair) st.cdr).car instanceof String))
       return super.scanForDefinitions(st, forms, defs, tr);
     Object name = p.car;
-    Type typeBinding = ClassType.make("gnu.mapping.Binding");
-    Declaration decl = defs.addDeclaration((String) name, typeBinding);
+    Type typeSymbol = ClassType.make("gnu.mapping.Symbol");
+    Declaration decl = defs.addDeclaration((String) name, typeSymbol);
     decl.setIndirectBinding(true);
-    st = tr.makePair(st, this, tr.makePair(p, decl, p.cdr));
+    st = Translator.makePair(st, this, Translator.makePair(p, decl, p.cdr));
     forms.addElement(st);
     return true;
   }
