@@ -64,9 +64,11 @@
 
 ;; Slib specific: (format #t "~q")
 
-(format #t "This implementation has~@[ no~] flonums ~
-            ~:[but no~;and~] complex numbers~%"
-	(not format:floats) format:complex-numbers)
+;; Kawa does have complex numbers, but does not support the
+;; non-standard ~i format specifier.
+;;(format #t "This implementation has~@[ no~] flonums ~
+;;            ~:[but no~;and~] complex numbers~%"
+;;	(not format:floats) format:complex-numbers)
 
 ; any object test
 
@@ -214,8 +216,8 @@ def")
 
 ")
 (test '("~|") (string slib:form-feed))
-(test '("~_~_~_") "   ")
-(test '("~3_") "   ")
+;; SLIB specific, conflicts with CommonLisp: (test '("~_~_~_") "   ") 
+;; SLIB specific, conflicts with CommonLisp: (test '("~3_") "   ")
 ;; SLIB specific, conflicts with CommonLisp: (test '("~/") (string slib:tab))
 ;; SLIB specific: (test '("~3/") (make-string 3 slib:tab))
 
@@ -323,6 +325,7 @@ def")
 (set! fail-expected "format:iobj-case-conv not implemented")
 (test `("~a" ,display) (string-upcase (format:iobj->str display)))
 (set! format:iobj-case-conv string-downcase)
+(set! fail-expected "format:iobj-case-conv not implemented")
 (test `("~s" ,display) (string-downcase (format:iobj->str display)))
 (set! format:iobj-case-conv string-capitalize)
 (set! fail-expected "format:iobj-case-conv not implemented")
@@ -485,6 +488,8 @@ def")
   (test '("~1f" 3.141) "3.")
   (test '("~f" 123.56789) "123.56789")
   (test '("~f" -314.0) "-314.0")
+  (test '("~f" 0.6) "0.6")
+  (test '("~f" 0.7) "0.7")
   (test '("~f" 1e4) "10000.0")
   (test '("~f" -1.23e10) "-12300000000.0")
   (test '("~f" 1e-4) "0.0001")
@@ -501,6 +506,8 @@ def")
  (format:floats
   (test '("~e" 3.14159) "3.14159E+0")
   (test '("~e" 0.00001234) "1.234E-5")
+  (test '("~e" 0.6) "6.0E-1")
+  (test '("~e" 0.7) "7.0E-1")
   (test '("~,,,0e" 0.00001234) "0.1234E-4")
   (test '("~,3e" 3.14159) "3.142E+0")
   (test '("~,3@e" 3.14159) "+3.142E+0")
@@ -634,6 +641,13 @@ def")
      ; number though format:complex is #f; this is why we put them in
      ; strings 
 
+(test '("{~:<[~3d]~:>}" (2 3)) "{([  2])}")
+(test '("{~:<[~3d]~:>}" 23) "{23}")
+(test '("~:<~W ~@_~:I~W ~:_~W~1I ~_~W~:>" (defun prod (x y) (* x y)))
+      "(defun prod (x y) (* x y))")
+(test '("~:<~W.~:I~W.~W~1I.~W~:>" (defun prod (x y) (* x y)))
+      "(defun.prod.(x y).(* x y))")
+
 ; inquiry test
 
 ;; SLIB specific: (test '("~:q") format:version)
